"""
   !--------------------------------------------------------------------------!
   ! LICENSE INFO:                                                            !
   !--------------------------------------------------------------------------!
   !    This file is part of localCIDER.                                      !
   !                                                                          !
   !    Version 0.1.9                                                         !
   !--------------------------------------------------------------------------!

   File Description:
   ================

   This is one of the key user interface/API files from which users (AKA you!)
   should use to access localCIDER.

   For a full description please see the documentation!

"""

from backend.sequence import Sequence
from backend.seqfileparser import SequenceFileParser
from backend.backendtools import status_message
from backend import plotting
from backend.localciderExceptions import SequenceException
from backend.localciderExceptions import SequenceComplexityException

class SequenceParameters:
    """
    A SequenceParameters object is the main object for examining sequence properties 
    in localCIDER. From this object a wide array of methods can be called on the
    object to explore a wide range of sequence properties.

    """

    def __init__(self, sequence="", sequenceFile=""):

        # provide the flexibility to submit either a sequence
        # file or an actual sequence as a string
        if sequence == "" and sequenceFile == "":
            raise SequenceException("Empty sequence/sequence file")

        # if the sequence isn't empty constuct a local
        # sequence object using the sequence
        if not sequence == "":

            # note we set the validate flag to check the sequence is valid and deal
            # with whitespace
            self.SeqObj = Sequence(sequence, validateSeq=True)
        else:
            parserMachine = SequenceFileParser()
            self.SeqObj = Sequence(parserMachine.parseSeqFile(sequenceFile))

    # ============================================ #
    # ============= SETTER FUNCTIONS ============= #

    #...................................................................................#
    def set_phosphosites(self, phosphosites):
        """
        Set putativly phosphorylated sites on your sequence.
        You can call this multiple times to update, though it
        will only ever add sites to the list. To clear sites
        use the clear_phosphosites() method.

        INPUT:
        --------------------------------------------------------------------------------
        phosphosites    |  list of site positions, indexed from 1


        OUTPUT:
        --------------------------------------------------------------------------------
        None, but the underlying Sequence object has its properties
        updated

        """

        # note that we do all the relevant data validation
        # in the calling function
        self.SeqObj.setPhosPhoSites(phosphosites)

    #...................................................................................#
    def clear_phosphosites(self):
        """
        Clears the putative phosphosites on the underlying
        sequence object. Useful if you want to reset the phosphostatus.

        OUTPUT:
        --------------------------------------------------------------------------------
        None, but the underlying Sequence object has its properties updated

        """

        self.SeqObj.clear_phosphosites()

    # ============================================ #
    # ============= GETTER FUNCTIONS ============= #

    #...................................................................................#
    def get_sequence(self):
        """
        Get the protein's primary amino acid sequence

        OUTPUT:
        --------------------------------------------------------------------------------
        Single string with the protein sequence

        """

        return self.SeqObj.seq

    #...................................................................................#
    def get_length(self):
        """
        Get the protein's amino acid sequence length

        OUTPUT:
        --------------------------------------------------------------------------------
        Integer equal to the sequence length

        """

        return len(self.SeqObj.seq)

    #...................................................................................#
    def get_mean_hydropathy(self):
        """
        Get a protein's mean hydropathy (a value between 0 and 9, where 0 is the least
        hydrophobic and 9 is the most). This is simply a re-based Kyte-Doolittle scale,
        which runs from 0 to 9 instead of from -4.5 to 4.5, as in the original paper.

        ********************************************************************************
        Ref: Kyte, J., & Doolittle, R. F. (1982). A simple method for displaying the
        hydropathic character of a protein. Journal of Molecular Biology, 157(1),
        105-132.
        ********************************************************************************

        OUTPUT:
        --------------------------------------------------------------------------------
        Float with the sequence's mean hydropathy

        """

        return self.SeqObj.meanHydropathy()

    #...................................................................................#
    def get_uversky_hydropathy(self):
        """
        Get a protein's mean hydropathy as defined by Uversky,
        using a normalized Kyte-Doolittle hydropathy table. Here, values range between
        0 and 1, with 0 being the least hydrophobic and 1 the most hydrophobic.

        OUTPUT:
        --------------------------------------------------------------------------------
        Float with the normalized Kyte-Doolittle hydropathy (between 0 and 1)

        """

        return self.SeqObj.uverskyHydropathy()

    #...................................................................................#
    def get_fraction_disorder_promoting(self):
        """
        Get a protein's fraction of residues which are considered '[D]isorder promoting'.

        For more details see the reference below;

        ********************************************************************************
        Ref: Campen A, Williams RM, Brown CJ, Meng J, Uversky VN, Dunker AK. (2008).
        TOP-IDP-scale: a new amino acid scale measuring propensity for intrinsic disorder.
        Protein Pept Lett. 15(9), 956-63.

        ********************************************************************************


        OUTPUT:
        --------------------------------------------------------------------------------
        Float with the fraction of disorder promoting residues

        """

        return self.SeqObj.fraction_disorder_promoting()

    #...................................................................................#
    def get_amino_acid_fractions(self):
        """
        Returns a dictionary with the fractions of each amino acid in your sequence

        OUTPUT:
        --------------------------------------------------------------------------------
        Dictionary of amino acids, where the keys are each of the 20 amino acids and the
        values represents the fraction of that amino acid

        """

        return self.SeqObj.amino_acid_fraction()

    #...................................................................................#
    def get_kappa(self):
        """
        Get the kappa value associated with a sequence.

        ********************************************************************************
        Ref: Das, R. K., & Pappu, R. V. (2013). Conformations of intrinsically disordered
        proteins are influenced by linear sequence distributions of oppositely
        charged residues. PNAS, 110(33), 13392-13397.
        ********************************************************************************

        OUTPUT:
        --------------------------------------------------------------------------------
        Float with the sequence's kappa value

        """

        return self.SeqObj.kappa()

    #...................................................................................#
    def get_Omega(self):
        """
        Get the Omega value associated with a sequence. Omega describes
        the patterning between charged/proline residues and all other residues

        ********************************************************************************
        Ref: Martin, E. W., Holehouse A. S.,  Pappu, R.V.  & Mittag, T. (2016). Sequence 
        determinants of the conformational properties of an intrinsically disordered 
        protein prior to and upon multi-site phosphorylation (submitted)
        ********************************************************************************

        OUTPUT:
        --------------------------------------------------------------------------------
        Float with the sequence's kappa-proline value

        """

        return self.SeqObj.kappa_proline()

    #...................................................................................#
    def get_Omega_sequence(self):
        """
        Get the 2-alphabet sequence used for calculating the Omega parameter
        as defined in Martin et al. R/K/D/E/P are represented as X and all other residues
        are O.

        ********************************************************************************
        Ref: Martin, E. W., Holehouse A. S.,  Pappu, R.V.  & Mittag, T. (2016). Sequence 
        determinants of the conformational properties of an intrinsically disordered 
        protein prior to and upon multi-site phosphorylation (submitted)
        ********************************************************************************

        OUTPUT:
        --------------------------------------------------------------------------------
        String giving the sequence in a reduced 2-letter alphabet.

        """

        return self.SeqObj.kappa_proline_seq()

    #...................................................................................#
    def get_kappa_X(self, grp1, grp2=None):
        """
        Get the user defined patterning paremeter, where residues are grouped into either
        two groups (grp1 and ALL other residues) or three groups (grp1, grp2, and ALL
        other residues).

        For context, kappa would be calulated using
        grp1 = ['E','D']
        grp2 = ['K','D']

        While for Omgega definition the grouping would be
        grp1 = ['P','E','D','K','R']
        (note grp2 is left =None)
                
        OUTPUT:
        --------------------------------------------------------------------------------
        Float with the user-defined patterning parameter

        """

        return self.SeqObj.kappa_X(grp1, grp2)

    #...................................................................................#
    def get_deltaMax(self):
        """
        Get the maximum delta value for a sequence of this composition. Note kappa is
        delta/deltaMax.

        OUTPUT:
        --------------------------------------------------------------------------------
        Float with the sequence's delta max (identical for all permutants)

        """

        return self.SeqObj.deltaMax()

    #...................................................................................#
    def get_delta(self):
        """
        Get the delta value for this specific sequence. Note kappa is delta/deltaMax.

        OUTPUT:
        --------------------------------------------------------------------------------
        Float with the sequence's delta max (will vary with permutants)
        """

        return self.SeqObj.delta()

    #...................................................................................#
    def get_countPos(self):
        """
        Get the number of positive residues in the sequence

        OUTPUT:
        --------------------------------------------------------------------------------
        Integer with number of positive residues in your sequence

        """

        return self.SeqObj.countPos()

    #...................................................................................#
    def get_countNeg(self):
        """
        Get the number of negative residues in the sequence

        OUTPUT:
        --------------------------------------------------------------------------------
        Integer with number of negative residues in your sequence

        """

        return self.SeqObj.countNeg()

    #...................................................................................#
    def get_countNeut(self):
        """
        Get the number of neutral residues in the sequence

        OUTPUT:
        --------------------------------------------------------------------------------
        Integer with number of neutral residues in your sequence

        """

        return self.SeqObj.countNeut()

    #...................................................................................#
    def get_fraction_positive(self):
        """
        Get the fraction of positively charged residues in the sequence

        OUTPUT:
        --------------------------------------------------------------------------------
        Float with the sequence's fraction of positively charged residues (F+)
        """

        return self.SeqObj.Fplus()

    #...................................................................................#
    def get_fraction_negative(self):
        """
        Get the fraction of negatively residues in the sequence

        OUTPUT
        --------------------------------------------------------------------------------:
        Float with the sequence's fraction of negatively charged residues (F-)

        """

        return self.SeqObj.Fminus()

    #...................................................................................#
    def get_FCR(self):
        """
        Get the fraction of charged residues in the sequence

        OUTPUT:
        --------------------------------------------------------------------------------
        Float with the sequence's fraction of charged residues

        """

        return self.SeqObj.FCR()

    #...................................................................................#
    def get_fraction_expanding(self):
        """
        Get the fraction of expanding residues in the sequence. We define 'expanding'
        residues as D/E/R/K/P. This will be the same as the FCR+fraction of proline

        OUTPUT:
        --------------------------------------------------------------------------------
        Float with the sequence's fraction of expanidng residues

        """

        return self.SeqObj.FER()

    #...................................................................................#
    def get_NCPR(self):
        """
        Get the net charge per residue of the sequence

        OUTPUT:
        --------------------------------------------------------------------------------
        Float with the sequence's net charge per residue

        """

        return self.SeqObj.NCPR()

    #...................................................................................#
    def get_mean_net_charge(self):
        """
        Get the absolute magnitude of the mean net charge

        OUTOUT:
        --------------------------------------------------------------------------------
        Float equal to the [absolute magnitude] of the mean net charge of the sequence

        """

        return self.SeqObj.mean_net_charge()

    #...................................................................................#
    def get_phasePlotRegion(self):
        """
        Returns the IDP diagram of states [REF 1] region based on the FCR/NCPR
        rules. Possible return values are;


            1 +-----------------------+---------------+
              |                     +                 |
              |                   +                   |
              |      4          +                     |
        F     |               +                       |
        R     |             +                         |
        A     |           +                           +
        C     |         +                           + |
        T     |       +                           +   |
        .     |     +            3              +     |
          0.5 |   +                           +       |
        N     | +                           +         |
        E     +                           +           |
        G     | +                       +             |
              + 2 +                   +               |
        R     | +   +               +       5         |
        E     |   + 2 +           +                   |
        S     |     +   +       +                     |
              | 1     + 2 +   +                       |
           0  +---------+---+-------------------------+

              0                  0.5                  1
                    FRACTION OF POSITIVE RESIDUES


        1) Weak polyampholytes and polyelectrolytes
        2) Intermediate (Janus) sequences
        3) Strong polyampholytes
        4 and 5) Strong polyelectrolytes

        OUTPUT:
        --------------------------------------------------------------------------------
        Returns an integer describing the region on the density of states
        diagram (above)

        """

        return self.SeqObj.phasePlotRegion()

    #...................................................................................#
    def get_phosphosites(self):
        """
        Function which returns a list of currently assigned
        phosphosites on your sequence.

        OUTPUT: Returns a list of integers where each
        integer represents a site currently defined as
        phosphorylatable. Such sites *must*, by definition
        by T/Y/S.

        OUTPUT:
        --------------------------------------------------------------------------------
        Returns a list of integers corresponding to the sites
        which are currently defined as being phosphorylatable based on
        user input

        """

        return self.SeqObj.get_phosphosites()

    #...................................................................................#
    def get_kappa_after_phosphorylation(self):
        """
        Function which recomputes kappa after complete
        phosphorylation based on the currently defined
        phosphosites.

        OUTPUT:
        --------------------------------------------------------------------------------
        returns a float corresponding to the sequence's kappa value if
        all the currently defined phosphosites were phosphorylated

        """

        if len(self.get_phosphosites()) == 0:
            status_message(
                "Be aware that there are no phosphosites currently set - getting 'naked' kappa")
        return self.SeqObj.kappa_at_maxPhos()

    #...................................................................................#
    def get_all_phosphorylatable_sites(self):
        """
        Function which returns a list of all the positions which *could* be
        phosphorylated (i.e. are T/S/Y). NOTE this does not use any kind of
        smart lookup, metadata, or analysis. It's literally, where are the Y/T/S
        residues.

        Note positions are returned as indexed from 1 (so you can feed these positions
        directly into the set_phosphosites function.

        OUTPUT:
        --------------------------------------------------------------------------------
        Returns a list of integers corresponding to S/T/Y positions in your sequence

        """

        return self.SeqObj.get_STY_residues()

    #...................................................................................#
    def get_full_phosphostatus_kappa_distribution(self):
        """
        This function calculates the kappa value of all possible phosphorylation
        states, given the defined phosphosites.

        This is computationally tractable for small numbers of phosphosites, but can
        rapidly become extremely expensive.


        OUTPUT:
        --------------------------------------------------------------------------------
        Returns a list of tuples, where each tuple corresponds to a unique phosphostate
        of the protein of interest. Each position within the tuple is defined as follows;

        0 - kappa of sequence
        1 - Fraction of positive residues (F+) (does not change)
        2 - Fraction of negative residues (F-)
        3 - Fraction of charged residues  (FCR)
        4 - Net Charge Per Residue        (NCRP)
        5 - Mean hydropathy
        6 - phosphostatus

        These are all self explanatory, with the exception of phosphostatus, which defines
        a tuple with a position for each phosphorylatable site, set to 0 if not phosphorylated
        and 1 if phosphorylated. As an example, if I had a protein with three phosphosites
        (S4,Y43,S105), the tuple for the fully unphosphorylated would be (0,0,0) and with
        Y43 phosphorylated would be (0,1,0)

        """

        # determine the number of calculations needed to run
        ncalcs = self.SeqObj.calculateNumberDifferentPhosphoStates()

        # print status message
        status_message(
            "Running exaustive kappa distribution analysis based on phosphorylation states")
        status_message(
            "This function will now make " +
            str(ncalcs) +
            " independent kappa calculations\nIf this is a big number you may want to investigate a subset of possible phosphosites or\nuse a Monte Carlo approach to subsample")

        return self.SeqObj.calculateKappaDistOfPhosphoStates()

    #...................................................................................#
    def get_phosphosequence(self):
        """
        Returns the sequence with phosphorylated sites set to E instead of S/Y/T
        """

        return self.SeqObj.get_phosphosequence()

    #...................................................................................#
    def get_PPII_propensity(self):
        """
        Get the overall sequence's PPII propensity as defined by Hilser et al [1] using
        the values reported in table 1 from [2]

        [1] - Elam WA, Schrank TP, Campagnolo AJ, Hilser VJ. Evolutionary 
        conservation of the polyproline II conformation surrounding intrinsically 
        disordered phosphorylation sites. 
        Protein Sci. 2013; 22: 405- 417. doi: 10.1002/pro.2217 PMID: 23341186

        [2] - Tomasso, M. E., Tarver, M. J., Devarajan, D. & Whitten, S. T. 
        Hydrodynamic Radii of Intrinsically Disordered Proteins Determined 
        from Experimental Polyproline II Propensities. 
        PLoS Comput. Biol. 12, e1004686 (2016).

        OUTPUT:
        --------------------------------------------------------------------------------
        Float with the sequence's f_{PPII, chain}

        """

        return self.SeqObj.FPPII_chain()


    # =============================================== #
    # =======  LINEAR SEQUENCE INFORMATION  ========= #
    #...................................................................................#
    def get_linear_NCPR(self, blobLen=5):
        """
        Returns a numpy vector of the net charge per residue (NCPR) as defined by a 
        sliding window. The first dimension contains the values and the second the associated
        index values along the sequence. A stepsize of 1 is always used.

        blobLen     | Sliding window size over which NCPR is calculated (default = 5
                      to match the default for kappa calculation)

        """

        return(self.SeqObj.linearDistOfNCPR(blobLen))


    #...................................................................................#
    def get_linear_FCR(self, blobLen=5):
        """
        Returns a 2D numpy vector of the fraction of charged residues (FCR) as defined by a 
        sliding window. The first dimension contains the values and the second the associated
        index values along the sequence. A stepsize of 1 is always used.
       
        blobLen     | Sliding window size over which FCR is calculated (default = 5
                      to match the default for kappa calculation)

        """

        return(self.SeqObj.linearDistOfFCR(blobLen))


    #...................................................................................#
    def get_linear_hydropathy(self, blobLen=5):
        """
        Returns a numpy vector of the Kyte-Doolitle hydropathy of a sequence as defined
        by a sliding window. The first dimension contains the values and the second the associated
        index values along the sequence. A stepsize of 1 is always used.

        blobLen     | Sliding window size over which the hydrophobicity is calculated (default = 5
                      to match the default for kappa calculation)

        """

        return(self.SeqObj.linearDistOfHydropathy(blobLen))


    #...................................................................................#
    def get_linear_sequence_composition(self, blobLen=5):
        """
        Returns a numpy matrix where each


        blobLen     | Sliding window size over which the hydrophobicity is calculated (default = 5
                      to match the default for kappa calculation)

        """


        return(self.SeqObj.linearCompositions(blobLen))
        

    # =============================================== #
    # ======= SEQUENCE COMPLEXITY FUNCTIONS ========= #

    #...................................................................................#
    def get_reduced_alphabet_sequence(self, alphabetSize=20, userAlphabet={}):
        """"
        Get your sequence converted into a lower resolution (reduced) alphabet. A set of reduced alphabets exist and are defined below, or the user can define their own alphabet.

        INPUT:
        --------------------------------------------------------------------------------
        alphabetSize  | Defines the size of the alphabet being used, where pre-defined
                        alphabets are then used based on the specific size. Those
                        pre-defined alphabets are defined below.

        userAlphabet  | Allows the user to define their own alphabet. The format here
                        is a dictionary where each key-value pair is amino-acid to translation.
                        This means you need a dictionary of length 20 where each amino acid
                        is mapped to another amino acid. This is kind of tedious, but it helps
                        avoid user-error where specific amino acids are missed.

        OUTPUT:
        --------------------------------------------------------------------------------
        Returns an amino acid squence which has been reduced down to a simple composition
        based on the defined alphabet. Note this returns the sequence only, not a
        SequenceParameters object.

        Predefined alphabets shown below - all except eleven are based on alphabets defined in
        the reference below.

        two      - [(LVIMCAGSTPFYW), (EDNQKRH)]
        three    - [(LVIMCAGSTP), (FYW), (EDNQKRH)]
        four     - [(LVIMC), (AGSTP), (FYW), (EDNQKRH)]
        five     - [(LVIMC), (ASGTP), (FYW), (EDNQ), (KRH)]
        six      - [(LVIM), (ASGT), (PHC), (FYW), (EDNQ), (KR)]
        eight    - [(LVIMC), (AG), (ST), (P), (FYW), (EDNQ), (KR), (H)]
        ten      - [(LVIM), (C), (A), (G), (ST), (P), (FYW), (EDNQ), (KR), (H)]
        eleven   - [(LVIM), (C), (A), (G), (ST), (P), (FYW), (ED), (NQ), (KR), (H)]
        twelve   - [(LVIM), (C), (A), (G), (ST), (P), (FY), (W), (EQ), (DN), (KR), (H)]
        fifteen  - [(LVIM), (C), (A), (G), (S), (T), (P), (FY), (W), (E), (Q), (D), (N), (KR), (H)]
        eighteen - [(LM), (VI), (C), (A), (G), (S), (T), (P), (F), (Y), (W), (E), (D), (N), (Q), (K), (R), (H)]
        twenty   - all twenty!

        REF: Murphy, L. R., Wallqvist, A., & Levy, R. M. (2000). Simplified amino acid alphabets for
        protein fold recognition and implications for folding. Protein Engineering, 13(3), 149-152.


        """
        return self.SeqObj.get_reducedAlphabetSequence(
            alphabetSize, userAlphabet)

    #...................................................................................#
    def get_linear_complexity(
            self,
            complexityType="WF",
            alphabetSize=20,
            userAlphabet={},
            blobLen=10,
            stepSize=1,
            wordSize=3):
        """
        Returns the linear sequence complexity as defined by complexityType. Optionally,
        the sequence complexity of a reduced complexity alphabet can be returned, where
        that reduced alphabet is defined by either the alphabetSize or the userAlphabet
        dictionary.

        INPUT:
        --------------------------------------------------------------------------------
        complexityType | Defines the complexity measure being employed. Is a string equal
                         to one of the opions described below;

                         WF  - Wooton-Federhen complexity [1]

                         LC  - Linguistic complexity      [2]

                         LZW - Lempel-Ziv-Welch           [3]


                         (Default = 'WF')

        alphabetSize   | Defines the size of the alphabet being used, where pre-defined
                         alphabets are then used based on the specific size. Those
                         pre-defined alphabets are defined below. (Default = 20, i.e. no
                         reduction in amino acid complexity)

        userAlphabet   | Allows the user to define their own alphabet. The format here
                         is a dictionary where each key-value pair is amino-acid to translation.
                         This means you need a dictionary of length 20 where each amino acid
                         is mapped to another amino acid. This is kind of tedious, but it helps
                         avoid user-error where specific amino acids are missed. (default=None)

        blobLen     | Sliding window size over which complexity is calculated (default=10)

        stepSize       | Size of steps taken as we define a new sliding window. Default is
                         1 (recommended to keep at 1)

        wordSize       | Relevant for linguistic complexity, ignored for other types. Default
                         is 3

        OUTPUT:
        --------------------------------------------------------------------------------
        Returns a vector of values corresponding to the sliding window complexity of the
        sequence, using the measure defined, and using the reduced alphabet complexity as
        defined

        Predefined alphabets shown below - all except eleven are based on alphabets defined in
        the reference below [4]

        two      - [(LVIMCAGSTPFYW), (EDNQKRH)]
        three    - [(LVIMCAGSTP), (FYW), (EDNQKRH)]
        four     - [(LVIMC), (AGSTP), (FYW), (EDNQKRH)]
        five     - [(LVIMC), (ASGTP), (FYW), (EDNQ), (KRH)]
        six      - [(LVIM), (ASGT), (PHC), (FYW), (EDNQ), (KR)]
        eight    - [(LVIMC), (AG), (ST), (P), (FYW), (EDNQ), (KR), (H)]
        ten      - [(LVIM), (C), (A), (G), (ST), (P), (FYW), (EDNQ), (KR), (H)]
        eleven   - [(LVIM), (C), (A), (G), (ST), (P), (FYW), (ED), (NQ), (KR), (H)]
        twelve   - [(LVIM), (C), (A), (G), (ST), (P), (FY), (W), (EQ), (DN), (KR), (H)]
        fifteen  - [(LVIM), (C), (A), (G), (S), (T), (P), (FY), (W), (E), (Q), (D), (N), (KR), (H)]
        eighteen - [(LM), (VI), (C), (A), (G), (S), (T), (P), (F), (Y), (W), (E), (D), (N), (Q), (K), (R), (H)]
        twenty   - all twenty!

        --------------------------------------------------------------------------------
        REFERENCES

        [1] Wootton, J. C., & Federhen, S. (1993). Statistics of local complexity in amino acid sequences
            and sequence databases. Computers & Chemistry, 17(2), 149-163.

        [2] Troyanskaya, O. G., Arbell, O., Koren, Y., Landau, G. M., & Bolshoy, A. (2002). Sequence
            complexity profiles of prokaryotic genomic sequences: a fast algorithm for calculating linguistic
            complexity. Bioinformatics , 18(5), 679 - 688.

        [3] Lempel, A.& Ziv, J. (1976). On the complexity of finite sequence. IEEE Trans. Inf. Theory,
            vol. IT-22, no. 1, 75-81.

        [4] Murphy, L. R., Wallqvist, A., & Levy, R. M. (2000). Simplified amino acid alphabets for
        protein fold recognition and implications for folding. Protein Engineering, 13(3), 149-152.

        """

        # set the allowed types of complexity here
        allowed_types = ('WF', 'LC', 'LZW')

        # provide case insensitivity
        try:
            complexityType = complexityType.upper()
        except AttributeError:
            pass

        # check fi the type passed is actually one of the ones we know about
        if complexityType not in allowed_types:
            raise SequenceComplexityException(
                "Complexity type %s is not a valid type - must be one of %s" %
                (complexityType, allowed_types))

        if complexityType == "WF":
            if not wordSize == 3:
                print "WARNING: Ignoring wordSize argument for Wooton-Federhen complexity"

            return self.SeqObj.get_linear_WF_complexity(
                alphabetSize, userAlphabet, blobLen, stepSize)

        if complexityType == "LZW":
            if not wordSize == 3:
                print "WARNING: Ignoring wordSize argument for LZW complexity"

            return self.SeqObj.get_linear_LZW_complexity(
                alphabetSize, userAlphabet, blobLen, stepSize)

        # coming soon - a new complexity measure...!
        if complexityType == "RHP":
            """
            if not wordSize == 3:
                print "WARNING: Ignoring wordSize argument for RHP complexity"

            return self.SeqObj.get_linear_RHP_complexity(
                alphabetSize, userAlphabet, blobLen, stepSize)
            """

        if complexityType == "LC":
            return self.SeqObj.get_linear_LC_complexity(
                alphabetSize, userAlphabet, blobLen, stepSize, wordSize)


    #...................................................................................#
    def show_linearComplexity(
            self,
            complexityType="WF",
            alphabetSize=20,
            userAlphabet={},
            blobLen=10,
            stepSize=1,
            wordSize=3,
            getFig=False):
        
        """
        Returns the linear sequence complexity as defined by complexityType. Optionally,
        the sequence complexity of a reduced complexity alphabet can be returned, where
        that reduced alphabet is defined by either the alphabetSize or the userAlphabet
        dictionary.

        INPUT:
        --------------------------------------------------------------------------------
        complexityType | Defines the complexity measure being employed. Is a string equal
                         to one of the opions described below;

                         WF  - Wooton-Federhen complexity [1]

                         LC  - Linguistic complexity      [2]

                         LZW - Lempel-Ziv-Welch           [3]

                         (Default = 'WF')

        alphabetSize   | Defines the size of the alphabet being used, where pre-defined
                         alphabets are then used based on the specific size. Those
                         pre-defined alphabets are defined below. (Default = 20, i.e. no
                         reduction in amino acid complexity)

        userAlphabet   | Allows the user to define their own alphabet. The format here
                         is a dictionary where each key-value pair is amino-acid to translation.
                         This means you need a dictionary of length 20 where each amino acid
                         is mapped to another amino acid. This is kind of tedious, but it helps
                         avoid user-error where specific amino acids are missed. (default=None)

        blobLen     | Sliding window size over which complexity is calculated (default=10)

        stepSize       | Size of steps taken as we define a new sliding window. Default is
                         1 (recommended to keep at 1)

        wordSize       | Relevant for linguistic complexity, ignored for other types. Default
                         is 3

        getFig         | Return the matplotlib figure object (for further customization)
                         (Default = False)

        OUTPUT:
        --------------------------------------------------------------------------------
        Returns a vector of values corresponding to the sliding window complexity of the
        sequence, using the measure defined, and using the reduced alphabet complexity as
        defined


        Predefined alphabets shown below - all except eleven are based on alphabets defined in
        the reference below [4]

        two      - [(LVIMCAGSTPFYW), (EDNQKRH)]
        three    - [(LVIMCAGSTP), (FYW), (EDNQKRH)]
        four     - [(LVIMC), (AGSTP), (FYW), (EDNQKRH)]
        five     - [(LVIMC), (ASGTP), (FYW), (EDNQ), (KRH)]
        six      - [(LVIM), (ASGT), (PHC), (FYW), (EDNQ), (KR)]
        eight    - [(LVIMC), (AG), (ST), (P), (FYW), (EDNQ), (KR), (H)]
        ten      - [(LVIM), (C), (A), (G), (ST), (P), (FYW), (EDNQ), (KR), (H)]
        eleven   - [(LVIM), (C), (A), (G), (ST), (P), (FYW), (ED), (NQ), (KR), (H)]
        twelve   - [(LVIM), (C), (A), (G), (ST), (P), (FY), (W), (EQ), (DN), (KR), (H)]
        fifteen  - [(LVIM), (C), (A), (G), (S), (T), (P), (FY), (W), (E), (Q), (D), (N), (KR), (H)]
        eighteen - [(LM), (VI), (C), (A), (G), (S), (T), (P), (F), (Y), (W), (E), (D), (N), (Q), (K), (R), (H)]
        twenty   - all twenty!

        --------------------------------------------------------------------------------
        REFERENCES

        [1] Wootton, J. C., & Federhen, S. (1993). Statistics of local complexity in amino acid sequences
            and sequence databases. Computers & Chemistry, 17(2), 149-163.

        [2] Troyanskaya, O. G., Arbell, O., Koren, Y., Landau, G. M., & Bolshoy, A. (2002). Sequence
            complexity profiles of prokaryotic genomic sequences: a fast algorithm for calculating linguistic
            complexity. Bioinformatics , 18(5), 679 - 688.

        [3] Lempel, A.& Ziv, J. (1976). On the complexity of finite sequence. IEEE Trans. Inf. Theory,
            vol. IT-22, no. 1, 75-81.

        [4] Murphy, L. R., Wallqvist, A., & Levy, R. M. (2000). Simplified amino acid alphabets for
        protein fold recognition and implications for folding. Protein Engineering, 13(3), 149-152.
        
        """

        # first get the complexity info
        linear_complexity_vector = self.get_linear_complexity(
            complexityType, 
            alphabetSize,
            userAlphabet,
            blobLen,
            stepSize,
            wordSize)

        # now generate the figure and either return the figure object or just show it
        if getFig:
            return plotting.show_linearComplexity(linear_complexity_vector, 
                                                  complexityType, 
                                                  len(self.SeqObj.seq), 
                                                  getFig)
        else:
            plotting.show_linearComplexity(linear_complexity_vector, 
                                           complexityType, 
                                           len(self.SeqObj.seq), 
                                           getFig)

    #...................................................................................#
    def save_linearComplexity(
            self,
            filename,
            complexityType="WF",
            alphabetSize=20,
            userAlphabet={},
            blobLen=10,
            stepSize=1,
            wordSize=3,
            saveFormat='png'):
        """
        Returns the linear sequence complexity as defined by complexityType. Optionally,
        the sequence complexity of a reduced complexity alphabet can be returned, where
        that reduced alphabet is defined by either the alphabetSize or the userAlphabet
        dictionary.

        INPUT:
        --------------------------------------------------------------------------------
        filename       | Name of a file to write to

        complexityType | Defines the complexity measure being employed. Is a string equal
                         to one of the opions described below;

                         WF  - Wooton-Federhen complexity [1]

                         LC  - Linguistic complexity      [2]

                         LZW - Lempel-Ziv-Welch           [3]

                         (Default = 'WF')

        alphabetSize   | Defines the size of the alphabet being used, where pre-defined
                         alphabets are then used based on the specific size. Those
                         pre-defined alphabets are defined below. (Default = 20, i.e. no
                         reduction in amino acid complexity)

        userAlphabet   | Allows the user to define their own alphabet. The format here
                         is a dictionary where each key-value pair is amino-acid to translation.
                         This means you need a dictionary of length 20 where each amino acid
                         is mapped to another amino acid. This is kind of tedious, but it helps
                         avoid user-error where specific amino acids are missed. (default=None)

        blobLen     | Sliding window size over which complexity is calculated (default=10)

        stepSize       | Size of steps taken as we define a new sliding window. Default is
                         1 (recommended to keep at 1)

        wordSize       | Relevant for linguistic complexity, ignored for other types. Default
                         is 3

        saveFormat     | Defines the file formal to save plots as. This parameter
                         is passed to matplotlibs savefig command which supports 
                         the following filetypes: emf, eps, pdf, png, ps, raw, 
                         rgba, svg, svgz. (DEFAULT = png)

        OUTPUT:
        --------------------------------------------------------------------------------
        Returns a vector of values corresponding to the sliding window complexity of the
        sequence, using the measure defined, and using the reduced alphabet complexity as
        defined


        Predefined alphabets shown below - all except eleven are based on alphabets defined in
        the reference below [4]

        two      - [(LVIMCAGSTPFYW), (EDNQKRH)]
        three    - [(LVIMCAGSTP), (FYW), (EDNQKRH)]
        four     - [(LVIMC), (AGSTP), (FYW), (EDNQKRH)]
        five     - [(LVIMC), (ASGTP), (FYW), (EDNQ), (KRH)]
        six      - [(LVIM), (ASGT), (PHC), (FYW), (EDNQ), (KR)]
        eight    - [(LVIMC), (AG), (ST), (P), (FYW), (EDNQ), (KR), (H)]
        ten      - [(LVIM), (C), (A), (G), (ST), (P), (FYW), (EDNQ), (KR), (H)]
        eleven   - [(LVIM), (C), (A), (G), (ST), (P), (FYW), (ED), (NQ), (KR), (H)]
        twelve   - [(LVIM), (C), (A), (G), (ST), (P), (FY), (W), (EQ), (DN), (KR), (H)]
        fifteen  - [(LVIM), (C), (A), (G), (S), (T), (P), (FY), (W), (E), (Q), (D), (N), (KR), (H)]
        eighteen - [(LM), (VI), (C), (A), (G), (S), (T), (P), (F), (Y), (W), (E), (D), (N), (Q), (K), (R), (H)]
        twenty   - all twenty!

        --------------------------------------------------------------------------------
        REFERENCES

        [1] Wootton, J. C., & Federhen, S. (1993). Statistics of local complexity in amino acid sequences
            and sequence databases. Computers & Chemistry, 17(2), 149-163.

        [2] Troyanskaya, O. G., Arbell, O., Koren, Y., Landau, G. M., & Bolshoy, A. (2002). Sequence
            complexity profiles of prokaryotic genomic sequences: a fast algorithm for calculating linguistic
            complexity. Bioinformatics , 18(5), 679 - 688.

        [3] Lempel, A.& Ziv, J. (1976). On the complexity of finite sequence. IEEE Trans. Inf. Theory,
            vol. IT-22, no. 1, 75-81.

        [4] Murphy, L. R., Wallqvist, A., & Levy, R. M. (2000). Simplified amino acid alphabets for
        protein fold recognition and implications for folding. Protein Engineering, 13(3), 149-152.


        """


            
        # first get the complexity info and build the correct vector
        linear_complexity_vector = self.get_linear_complexity(
            complexityType, 
            alphabetSize,
            userAlphabet,
            blobLen,
            stepSize,
            wordSize)

        # now generate and save the relevant figure
        plotting.save_linearComplexity(linear_complexity_vector, 
                                       complexityType, 
                                       len(self.SeqObj.seq), 
                                       filename,
                                       saveFormat)


        

        
        
    # ============================================ #
    # ======= PLOTTING DIAGRAM FUNCTIONS ========= #
    #...................................................................................#
    def show_phaseDiagramPlot(
            self,
            label="",
            title="Diagram of states",
            legendOn=True,
            xLim=1,
            yLim=1,
            fontSize=10,
            getFig=False):
        """
        Generates the Pappu-Das phase diagram (diagram of states), places
        this sequence on that plot, and creates it on the screen

        INPUT:
        --------------------------------------------------------------------------------
        label     | A label for the point on the phase diagram
        title     | Plot title (DEFAULT = 'Diagram of states')
        legendOn  | Boolean for if the figure legend should be displayed or not
        xLim      | Max value for the x axis (fract. positive charge) (DEFAULT = 1)
        yLim      | Max value for the y axis (fract. negative charge) (DEFAULT = 1)
        fontSize  | Size of font for label (DEFAULT = 10)
        getFig    | Returns a matplotlib figure object instead of simply displaying the
                  | plot on the screen (DEFAULT = False)


        OUTPUT:
        --------------------------------------------------------------------------------
        If the argument getFig is False (which it is by default) then the Uversky plot appears on
        the screen. If getFig is set to True then the function returns a matplotlib plt object, which
        can be further manipulated.

        """

        if getFig:
            return plotting.show_single_phasePlot(
                self.get_fraction_positive(),
                self.get_fraction_negative(),
                label,
                legendOn,
                xLim,
                yLim,
                fontSize,
                getFig)
        else:
            plotting.show_single_phasePlot(
                self.get_fraction_positive(),
                self.get_fraction_negative(),
                label,
                title,
                legendOn,
                xLim,
                yLim,
                fontSize,
                getFig)

    #...................................................................................#
    def save_phaseDiagramPlot(
            self,
            filename,
            label="",
            title="Diagram of states",
            legendOn=True,
            xLim=1,
            yLim=1,
            fontSize=10,
            saveFormat='png'):
        """
        Generates the Pappu-Das phase diagram (diagram of states), places
        this sequence on that plot, and saves it at the <filename> location

        INPUT:
        --------------------------------------------------------------------------------
        filename  | Writeable filename

        label      | A label for the point on the phase diagram
        title      | Plot title (DEFAULT = 'Diagram of states')
        legendOn   | Boolean for if the figure legend should be displayed or not
        xLim       | Max value for the x axis (fract. positive charge) (DEFAULT = 1)
        yLim       | Max value for the y axis (fract. negative charge) (DEFAULT = 1)
        fontSize   | Size of font for label (DEFAULT = 10)
        saveFormat | Defines the file formal to save plots as. This parameter
                     is passed to matplotlibs savefig command which supports 
                     the following filetypes: emf, eps, pdf, png, ps, raw, 
                     rgba, svg, svgz. (DEFAULT = png)



        OUTPUT:
        --------------------------------------------------------------------------------
        Nothing, but creates a .png file at the filename location

        """

        plotting.save_single_phasePlot(
            self.get_fraction_positive(),
            self.get_fraction_negative(),
            filename,
            label,
            title,
            legendOn,
            xLim,
            yLim,
            fontSize,
            saveFormat)

    #...................................................................................#
    def show_uverskyPlot(
            self,
            label="",
            title="Uversky plot",
            legendOn=True,
            xLim=1,
            yLim=1,
            fontSize=10,
            getFig=False):
        """
        Generates the Uversky phase diagram (hydropathy vs NCPR), places
        this sequence on that plot, and renders it on the screen

        INPUT:
        --------------------------------------------------------------------------------
        label     | A label for the point on the phase diagram
        title     | Plot title (DEFAULT = 'Uversky plot')
        legendOn  | Boolean for if the figure legend should be displayed or not
        xLim      | Max value for the x axis (mean net charge) (DEFAULT = 1)
        yLim      | Max value for the y axis (hydropathy) (DEFAULT = 1)
        fontSize  | Size of font for label (DEFAULT = 10)
        getFig    | Returns a matplotlib figure object instead of simply displaying the
                    plot on the screen (DEFAULT = False)


        OUTPUT:
        --------------------------------------------------------------------------------
        If the argument getFig is False (which it is by default) then the Uversky plot appears on
        the screen. If getFig is set to True then the function returns a matplotlib plt object, which
        can be further manipulated.


        """
        if getFig:
            return plotting.show_single_uverskyPlot(
                self.get_uversky_hydropathy(),
                self.get_mean_net_charge(),
                label,
                title,
                legendOn,
                xLim,
                yLim,
                fontSize,
                getFig)
        else:
            plotting.show_single_uverskyPlot(
                self.get_uversky_hydropathy(),
                self.get_mean_net_charge(),
                label,
                title,
                legendOn,
                xLim,
                yLim,
                fontSize,
                getFig)

    #...................................................................................#
    def save_uverskyPlot(
            self,
            filename,
            label="",
            title="Uversky plot",
            legendOn=True,
            xLim=1,
            yLim=1,
            fontSize=10,
            saveFormat='png'):
        
        """
        Generates the Uversky phase diagram (hydropathy vs NCPR), places
        this sequence on that plot, and saves it at the <filename> location

        INPUT:
        --------------------------------------------------------------------------------
        filename  | A writeable filename

        label      | A label for the point on the phase diagram
        title      | Plot title (DEFAULT = 'Uversky plot')
        legendOn   | Boolean for if the figure legend should be displayed or not
        xLim       | Max value for the x axis (mean net charge) (DEFAULT = 1)
        yLim       | Max value for the y axis (hydropathy) (DEFAULT = 1)
        fontSize   | Size of font for label (DEFAULT = 10)
        saveFormat | Defines the file formal to save plots as. This parameter
                     is passed to matplotlibs savefig command which supports 
                     the following filetypes: emf, eps, pdf, png, ps, raw, 
                     rgba, svg, svgz. (DEFAULT = png)



        OUTPUT:
        --------------------------------------------------------------------------------
        If the argument getFig is False (which it is by default) then the Uversky plot appears on
        the screen. If getFig is set to True then the function returns a matplotlib plt object, which
        can be further manipulated.

        """

        plotting.save_single_uverskyPlot(
            self.get_uversky_hydropathy(),
            self.get_mean_net_charge(),
            filename,
            label,
            title,
            legendOn,
            xLim,
            yLim,
            fontSize,
            saveFormat)

    #...................................................................................#
    def save_linearNCPR(self, filename, blobLen=5, saveFormat='png'):
        """
        Generates a plot of how the NCPR (net charge per residue) changes as we move
        along the linear amino acid sequence in blobLen size steps. This uses a sliding window
        approach and calculates the average within that window.

        INPUT:
        --------------------------------------------------------------------------------
        filename   | Name of the file to write
        bloblen    | Set the windowsize (DEFAULT = 5)
        saveFormat | Defines the file formal to save plots as. This parameter
                     is passed to matplotlibs savefig command which supports 
                     the following filetypes: emf, eps, pdf, png, ps, raw, 
                     rgba, svg, svgz. (DEFAULT = png)


        OUTPUT:
        --------------------------------------------------------------------------------
        Nothing, but creates a saved file at the filename location.


        """

        plotting.save_linearplot(
            plotting.build_NCPR_plot,
            self.SeqObj,
            blobLen,
            filename,
            saveFormat)

    #...................................................................................#
    def save_linearFCR(self, filename, blobLen=5, saveFormat='png'):
        """
        Generates a plot of how the FCR (fraction of charged residues) changes as we move
        along the linear amino acid sequence in blobLen size steps. This uses a sliding window
        approach and calculates the average within that window.

        INPUT:
        --------------------------------------------------------------------------------
        filename   | Name of the file to write
        bloblen    | Set the windowsize (DEFAULT = 5)
        saveFormat | Defines the file formal to save plots as. This parameter
                     is passed to matplotlibs savefig command which supports 
                     the following filetypes: emf, eps, pdf, png, ps, raw, 
                     rgba, svg, svgz. (DEFAULT = png)



        OUTPUT:
        --------------------------------------------------------------------------------
        Nothing, but creates a .png file at the filename location


        """

        plotting.save_linearplot(
            plotting.build_FCR_plot,
            self.SeqObj,
            blobLen,
            filename,
            saveFormat)

    #...................................................................................#
    def save_linearSigma(self, filename, blobLen=5, saveFormat='png'):
        """
        Generates a plot of how the Sigma parameter changes as we move
        along the linear amino acid sequence in blobLen size steps. This uses a sliding window
        approach and calculates the average within that window.

        Recall that sigma is defined as

        NCPR^2/FCR (net charge per residue squared divided by the fraction of charged residues)

        INPUT:
        --------------------------------------------------------------------------------
        filename   | Name of the file to write
        bloblen    | Set the windowsize (DEFAULT = 5)
        saveFormat | Defines the file formal to save plots as. This parameter
                     is passed to matplotlibs savefig command which supports 
                     the following filetypes: emf, eps, pdf, png, ps, raw, 
                     rgba, svg, svgz. (DEFAULT = png)
        

        OUTPUT:
        --------------------------------------------------------------------------------
        Nothing, but creates a .png file at the filename location

        """

        plotting.save_linearplot(
            plotting.build_sigma_plot,
            self.SeqObj,
            blobLen,
            filename,
            saveFormat)

    #...................................................................................#
    def save_linearHydropathy(self, filename, blobLen=5, saveFormat='png'):
        """
        Generates a plot of how the mean hydropathy changes as we move
        along the linear amino acid sequence in blobLen size steps. This uses a sliding window
        approach and calculates the average within that window.

        Hydropathy here is calculated using a NORMALIZED Kyte-Doolittle scale, where 1 is
        the most hydrophobic and 0 the least.

        INPUT:
        --------------------------------------------------------------------------------
        filename   | Name of the file to write
        bloblen    | Set the windowsize (DEFAULT = 5)
        saveFormat | Defines the file formal to save plots as. This parameter
                     is passed to matplotlibs savefig command which supports 
                     the following filetypes: emf, eps, pdf, png, ps, raw, 
                     rgba, svg, svgz. (DEFAULT = png)



        OUTPUT:
        --------------------------------------------------------------------------------
        Nothing, but creates a .png file at the filename location

        """

        plotting.save_linearplot(
            plotting.build_hydropathy_plot,
            self.SeqObj,
            blobLen,
            filename,
            saveFormat)


    #...................................................................................#
    def save_linearComposition(self, filename, blobLen=5, saveFormat='png', title='', plot_data=False):
        """
        Generates a plot that for the standard grouping of amino acids plots the local
        linear density of those groups along the sequence. Density is calculated using a
        sliding window, and the data is then fit to a Univariate cubic spline fit which
        smoothes out a lot of the noise associated with a sliding-window style analysis.

        INPUT:
        --------------------------------------------------------------------------------
        filename   | Name of the file to write
        blobLen    | Set the windowsize (DEFAULT = 5)
        saveFormat | Defines the file formal to save plots as. This parameter
                     is passed to matplotlibs savefig command which supports 
                     the following filetypes: emf, eps, pdf, png, ps, raw, 
                     rgba, svg, svgz. (DEFAULT = png)
        title      | Set the title for the figure (useful when generating many
                     profiles for different proteins
        plot_data  | [DEFAULT = False] will overlay the raw data on top of the 
                     cubic spline interpolation. This can be useful to check that 
                     the blobLen selected is actually capturing the relevant 
                     features in the underlying data.
        



        OUTPUT:
        --------------------------------------------------------------------------------
        Nothing, but creates a .png file at the filename location

        """

        out_tuple = self.get_linear_sequence_composition(blobLen)
        residues = out_tuple[0]
        density  = out_tuple[1]

        colors = ['red',  'blue', 'brown', 'green',  'black',  'orange', 'purple']
        names  = ['E/D.', 'R/K',  'E/D/R/K',  'Q/N/S/T/G/H',  'I/L/V/M', 'F/Y/W.' ,  'P']
        line_thickness = [2.5, 2.5, 3.5, 3.5, 3.5, 3.5, 3.5]

        plotting.save_local_composition_plot(residues, density, colors, names, filename, saveFormat, line_thickness=line_thickness, title=title, plot_data=plot_data)


    #...................................................................................#
    def show_linearNCPR(self, blobLen=5, getFig=False):
        """
        Generates a plot of how the NCPR (net charge per residue) changes as we move
        along the linear amino acid sequence in blobLen size steps. This uses a sliding window
        approach and calculates the average within that window.

        INPUT:
        --------------------------------------------------------------------------------
        bloblen  | Set the windowsize (DEFAULT = 5)
        getFig   | Do you want to get the matplotlib figure object (DEFAULY = FALSE)


        OUTPUT:
        --------------------------------------------------------------------------------
        Nothing, but the plot is displayed on screen

        """

        return plotting.show_linearplot(
            plotting.build_NCPR_plot, self.SeqObj, blobLen, getFig)

    #...................................................................................#
    def show_linearFCR(self, blobLen=5, getFig=False):
        """
        Generates a plot of how the FCR (fraction of charged residues) changes as we move
        along the linear amino acid sequence in blobLen size steps. This uses a sliding window
        approach and calculates the average within that window.

        INPUT:
        --------------------------------------------------------------------------------
        bloblen  | Set the windowsize (DEFAULT = 5)
        getFig   | Do you want to get the matplotlib figure object (DEFAULY = FALSE)


        OUTPUT:
        --------------------------------------------------------------------------------
        Nothing, but the plot is displayed on screen

        """

        return plotting.show_linearplot(
            plotting.build_FCR_plot, self.SeqObj, blobLen, getFig)

    #...................................................................................#
    def show_linearSigma(self, blobLen=5, getFig=False):
        """
        Generates a plot of how the sigma parameter changes as we move
        along the linear amino acid sequence in blobLen size steps. This uses a sliding window
        approach and calculates the average within that window.

        Recall that sigma is defined as

        NCPR^2/FCR (net charge per residue squared divided by the fraction of charged residues)

        INPUT:
        --------------------------------------------------------------------------------
        bloblen  | Set the windowsize (DEFAULT = 5)
        getFig   | Do you want to get the matplotlib figure object (DEFAULY = FALSE)


        OUTPUT:
        --------------------------------------------------------------------------------
        Nothing, but the plot is displayed on screen

        """

        return plotting.show_linearplot(
            plotting.build_sigma_plot, self.SeqObj, blobLen, getFig)

    #...................................................................................#
    def show_linearHydropathy(self, blobLen=5, getFig=False):
        """
        Generates a plot of how the mean hydropathy changes as we move
        along the linear amino acid sequence in blobLen size steps. This uses a sliding window
        approach and calculates the average within that window.

        Hydropathy here is calculated using a NORMALIZED Kyte-Doolittle scale, where 1 is
        the most hydrophobic and 0 the least.

        INPUT:
        --------------------------------------------------------------------------------
        bloblen  | Set the windowsize (DEFAULT = 5)
        getFig   | Do you want to get the matplotlib figure object (DEFAULY = FALSE)


        OUTPUT:
        --------------------------------------------------------------------------------
        Nothing, but the plot is displayed on screen

        """

        return plotting.show_linearplot(
            plotting.build_hydropathy_plot, self.SeqObj, blobLen, getFig)



    # ============================================== #
    # ============ FORMATTING FUNCTIONS ============ #

    #...................................................................................#
    def set_HTMLColorResiduePalette(self, colorDict):
        """
        Allows the user to define the colormapping used for the get_HTMLColorString.

        The input parameter is a dictionary which contains amino acid single letter codes
        as keys and colours as values. The colours must be one of the 17 standard HTML
        color names. These are;

        aqua, black, blue, fuchsia, gray, green, lime, maroon, navy, olive, orange,
        purple, red, silver, teal, white, and yellow.

        By default, polar residues are green, negative residues red, positive blue,
        proline fuschia and all others are black.

        The function carrys out checking to ensure that
        1) Colors are valid (case insensitive)
        2) All 20 amino acids are accounted for

        No return value is provided, but an exception is raised if the operation
        cannot be completed.

        """
        self.SeqObj.set_HTMLColorResiduePalette(colorDict)


    #...................................................................................#
    def get_HTMLColorString(self):
        """
        Returns a string which contains a correctly formatted HTML string of the sequence
        defined with specific coloring.

        The sequence is divided into 10 residue blocks, with a newline (<br>) every
        50 residues.

        """
        return self.SeqObj.get_HTMLColorString()

    # ============================================ #
    # ============ IMPLICIT FUNCTIONS ============ #

    #...................................................................................#
    def __unicode__(self):
        """ Returns the sequences """
        return "SequenceParameter [len=" + \
            str(len(self.SeqObj.seq)) + "], [seq='" + self.SeqObj.seq + "']"


    def __repr__(self):
        return "["+hex(id(self)) + "]: " + self.__unicode__()

    #...................................................................................#
    def __str__(self):
        """ Returns the sequences """
        return self.__unicode__()


    #...................................................................................#
    def __len__(self):
        """ Returns the sequence length """
        return len(self.SeqObj.seq)



""" 
   !--------------------------------------------------------------------------!
   ! LICENSE INFO:                                                            !
   !--------------------------------------------------------------------------!
   !    This file is part of localCIDER.                                      !
   !                                                                          !
   !    Version 0.1.9                                                         !
   !                                                                          !
   !    Copyright (C) 2014 - 2016                                             !
   !    The localCIDER development team (current and former contributors)     !
   !    Alex Holehouse, James Ahad, Rahul K. Das.                             !
   !                                                                          !
   !    localCIDER was developed in the lab of Rohit Pappu at Washington      !
   !    University in St. Louis. Please see the website for citation          !
   !    information:                                                          !
   !                                                                          !
   !    http://pappulab.github.io/localCIDER/                                 !
   !                                                                          !
   !    For more information please see the Pappu lab website:                !
   !                                                                          !
   !    http://pappulab.wustl.edu/                                            !
   !                                                                          !
   !    localCIDER is free software: you can redistribute it and/or modify    !
   !    it under the terms of the GNU General Public License as published by  !
   !    the Free Software Foundation, either version 3 of the License, or     !
   !    (at your option) any later version.                                   !
   !                                                                          !
   !    localCIDER is distributed in the hope that it will be useful,         !
   !    but WITHOUT ANY WARRANTY; without even the implied warranty of        !
   !    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         !
   !    GNU General Public License for more details.                          !
   !                                                                          !
   !    You should have received a copy of the GNU General Public License     !
   !    along with localCIDER.  If not, see <http://www.gnu.org/licenses/>.   !
   !--------------------------------------------------------------------------!
   ! AUTHORSHIP INFO:                                                         !
   !--------------------------------------------------------------------------!
   !                                                                          !
   ! MAIN AUTHOR:    Alex Holehouse                                           !
   !                                                                          !
   !--------------------------------------------------------------------------!

   Stand alone program to compute the maximally complex sequences associated with
   a string of length *n* made of a 20 letter alphabet. Note that while proteins
   have some bias towards particular amino acids (i.e. few Trp, lots of Leu) the
   maximally complexity sequence is a theoretical max which selects uniformly from
   the 20 amino acids. This is a distinction related to how we define complexity -
   i.e. here complexity is defined in an absolute entropic sense and not in a
   way biased towards the maximally complex sequence seen in nature. In practice
   this probably means that you will never see a natural sequence with a complexity  
   of 1, so one pratical post-processing approach might be to normalize sequences
   by the most complex natural sequence you observe to get a sense of relative 
   complexity.

   We felt that by providing a raw, entropic/information theoretic measure of 
   complexity and then suggesting that users use that in whatever means they 
   want is a much safer way, rather than assuming some biasing will apply to
   all users' needs.

   This isn't really part of localCIDER, but may be useful for theoretical calculatoins
   associated with complexity. Be wary of depleting the entropy pool, though!
   
"""


import random
import zlib

NUM_ITERATIONS = 200000
MAX_LENGTH     = 1000+1
LETTERS        = "qwertyipasdfghklcvnm"


# build a random flat sequence
maxComp=-1
maxCompDict={}
CONVERGENCE_TARGET=600
# convergence test
for i in xrange(0,0):
    maxComp=-1
    maxCompDict={}

    # seed the random number generator!
    random.seed()

    for convergenceIteration in xrange(1,200000):
        length=MAX_LENGTH

        # build random sequence
        seq=""
        for pos in xrange(0,length):
            seq=seq+random.choice(LETTERS)
        
        compressed = len(zlib.compress(seq))
        
        # update
        if maxComp < compressed:
            #print "Updatding at iter %i with %i" % (convergenceIteration,compressed)
            maxCompDict[compressed] = convergenceIteration
            maxComp = compressed

    #print maxCompDict.keys()

    if CONVERGENCE_TARGET in maxCompDict.keys():
        print "Found max after %i" % maxCompDict[CONVERGENCE_TARGET]

        
        with open("convergenceDist.dat","a") as fh:
            fh.write("%i \n" % maxCompDict[CONVERGENCE_TARGET])
    else:
        print "Did not find max"
        with open("convergenceDist.dat","a") as fh:
            fh.write("-1\n")


    
maxComplexity = {}

# for each sequence of some length between
# 5 and MAX_LENGTH
for length in xrange(5,MAX_LENGTH):

    print "On sequences of length %i " % length

    # intialize 
    maxComplexity[length] = -1

    for iteration in xrange(0, NUM_ITERATIONS):
        
        # bulild flat, random string
        seq=""
        for pos in xrange(0,length):
            seq=seq+random.choice(LETTERS)
        
        # determine sequence complexity as the length of the
        # compressed string
        compressed = len(zlib.compress(seq))
        
        # update
        if maxComplexity[length] < compressed:
            maxComplexity[length] = compressed


# once complete write to a csv

with open("sequence_complexity.dat", "w") as fh:
    # write each max complexity to file!
    for length in xrange(5, MAX_LENGTH):
        fh.write("%i \t %i\n" % (length, maxComplexity[length]))
        
print "Complexity done for sequences between 5 and %i (using %i iterations per sequence)" % (MAX_LENGTH, NUM_ITERATIONS)


    
    




"""
   !--------------------------------------------------------------------------!
   ! LICENSE INFO:                                                            !
   !--------------------------------------------------------------------------!
   !    This file is part of localCIDER.                                      !
   !                                                                          !
   !    Version 0.1.9                                                         !
   !                                                                          !
   !    Copyright (C) 2014 - 2016                                             !
   !    The localCIDER development team (current and former contributors)     !
   !    Alex Holehouse, James Ahad, Rahul K. Das.                             !
   !                                                                          !
   !    localCIDER was developed in the lab of Rohit Pappu at Washington      !
   !    University in St. Louis. Please see the website for citation          !
   !    information:                                                          !
   !                                                                          !
   !    http://pappulab.github.io/localCIDER/                                 !
   !                                                                          !
   !    For more information please see the Pappu lab website:                !
   !                                                                          !
   !    http://pappulab.wustl.edu/                                            !
   !                                                                          !
   !    localCIDER is free software: you can redistribute it and/or modify    !
   !    it under the terms of the GNU General Public License as published by  !
   !    the Free Software Foundation, either version 3 of the License, or     !
   !    (at your option) any later version.                                   !
   !                                                                          !
   !    localCIDER is distributed in the hope that it will be useful,         !
   !    but WITHOUT ANY WARRANTY; without even the implied warranty of        !
   !    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         !
   !    GNU General Public License for more details.                          !
   !                                                                          !
   !    You should have received a copy of the GNU General Public License     !
   !    along with localCIDER.  If not, see <http://www.gnu.org/licenses/>.   !
   !--------------------------------------------------------------------------!
   ! AUTHORSHIP INFO:                                                         !
   !--------------------------------------------------------------------------!
   !                                                                          !
   ! MAIN AUTHOR:   James Ahad and Alex Holehouse                             !
   !                                                                          !
   !--------------------------------------------------------------------------!


   File Description:
   ================

   Plotting is the backened for all localCIDER's plotting functionality.

   As with all functions in backend, nothing should be called directly, but in this case
   through the localcider.plots API module.

   Given plots are their own stateless function the plotting module is purely
   functional, no classes or state is maintained.

   For adding additional plotting features please see the paradigm defined with
   the linear plots - specifically using a hierarchical approach which passes
   functions. This makes for much cleaner code and reduces maintainance.


"""
import os
import matplotlib
import matplotlib.pyplot as plt
from matplotlib.font_manager import FontProperties
import numpy as np
from scipy.interpolate import interp1d
from scipy.interpolate import UnivariateSpline


from sequence import Sequence
from backendtools import verifyType
from localciderExceptions import PlottingException

#...................................................................................#


def show_single_phasePlot(
        fp,
        fn,
        label="",
        title="Diagram of states",
        legendOn=True,
        xLim=1,
        yLim=1,
        fontSize=10,
        getFig=False):
    """
    Display a single-sequence Das-Pappu phase diagram plot on the screen
    """

    phaseplot_validate(fp, fn)

    # Create a plotting object
    initial_plottingObject = single_plot(fp, fn, label, fontSize)
    finalized_plottingObject = finalize_DasPappu(
        initial_plottingObject, legendOn, title, xLim, yLim)

    # show the plot, or return the matplotlib fig object
    if getFig:
        return finalized_plottingObject
    else:
        finalized_plottingObject.show()


#...................................................................................#
def save_single_phasePlot(
        fp,
        fn,
        filename,
        label="",
        title="Diagram of states",
        legendOn=True,
        xLim=1,
        yLim=1,
        fontSize=10,
        saveFormat='pdf'):
    """
    Save a single-sequence Das-Pappu phase diagram to file
    """

    phaseplot_validate(fp, fn)

    # Delete any file in the filename location
    if(os.path.exists(filename)):
        os.remove(filename)

     # Create a plotting object
    initial_plottingObject = single_plot(fp, fn, label, fontSize)
    finalized_plottingObject = finalize_DasPappu(
        initial_plottingObject, legendOn, title, xLim, yLim)

    # save the plot and then close the matloblib plotting object
    # note we default to 200 dpi for PNG files - this value
    # is hardcoded
    if saveFormat == 'png':
        finalized_plottingObject.savefig(filename, format=saveFormat, dpi=200)
    else:
        finalized_plottingObject.savefig(filename, format=saveFormat)

    finalized_plottingObject.close()


#...................................................................................#
def show_multiple_phasePlot(
        fp_list,
        fn_list,
        label=[],
        title="Diagram of states",
        legendOn=True,
        xLim=1,
        yLim=1,
        fontSize=10,
        getFig=False):
    """
    Display multiple-sequences on a Das-Pappu phase diagram plot on the screen
    """

    # validate the various points
    for fp, fn in zip(fp_list, fn_list):
        phaseplot_validate(fp, fn)

    # Create a plotting object
    initial_plottingObject = multiple_plot(fp_list, fn_list, label, fontSize)
    finalized_plottingObject = finalize_DasPappu(
        initial_plottingObject, legendOn, title, xLim, yLim)

    # show the plot, or return the matplotlib fig object
    if getFig:
        return finalized_plottingObject
    else:
        finalized_plottingObject.show()


#...................................................................................#
def save_multiple_phasePlot(
        fp_list,
        fn_list,
        filename,
        label=[],
        title="Diagram of states",
        legendOn=True,
        xLim=1,
        yLim=1,
        fontSize=10,
        saveFormat='png'):
    """
    Save multiple-sequences on a Das-Pappu phase diagram to file
    """

    # validate the various points
    for fp, fn in zip(fp_list, fn_list):
        phaseplot_validate(fp, fn)

    # Create a plotting object
    initial_plottingObject = multiple_plot(fp_list, fn_list, label, fontSize)
    finalized_plottingObject = finalize_DasPappu(
        initial_plottingObject, legendOn, title, xLim, yLim)

    # save the plot and then close the matloblib plotting object
    # save the plot and then close the matloblib plotting object
    # note we default to 200 dpi for PNG files - this value
    # is hardcoded
    if saveFormat == 'png':
        finalized_plottingObject.savefig(filename, format=saveFormat, dpi=200)
    else:
        finalized_plottingObject.savefig(filename, format=saveFormat)

    
    finalized_plottingObject.close()


# =========================
###
# UVERSKY PLOT FUNCTIONS
###
# =========================
#...................................................................................#
def show_single_uverskyPlot(
        hydropathy,
        mean_net_charge,
        label="",
        title="Uversky plot",
        legendOn=True,
        xLim=1,
        yLim=1,
        fontSize=10,
        getFig=False):

    # Create a plotting object
    initial_plottingObject = single_plot(
        mean_net_charge, hydropathy, label, fontSize)
    finalized_plottingObject = finalize_uversky(
        initial_plottingObject, legendOn, title, xLim, yLim)

    # show the plot, or return the matplotlib fig object
    if getFig:
        return finalized_plottingObject
    else:
        finalized_plottingObject.show()


#...................................................................................#
def save_single_uverskyPlot(
        hydropathy,
        mean_net_charge,
        filename,
        label="",
        title="Uversky plot",
        legendOn=True,
        xLim=1,
        yLim=1,
        fontSize=10,
        saveFormat='png'):
    """
    Function to save a single point on a Uversky plots

    """

    # Delete any file in the filename location
    if(os.path.exists(filename)):
        os.remove(filename)

     # Create a plotting object
    initial_plottingObject = single_plot(
        mean_net_charge, hydropathy, label, fontSize)
    finalized_plottingObject = finalize_uversky(
        initial_plottingObject, legendOn, title, xLim, yLim)

    # save the plot and then close the matloblib plotting object
    if saveFormat == 'png':
        finalized_plottingObject.savefig(filename, format=saveFormat, dpi=200)
    else:
        finalized_plottingObject.savefig(filename, format=saveFormat)

    finalized_plottingObject.close()


#...................................................................................#
def show_multiple_uverskyPlot(
        hydropathy_list,
        mean_net_charge_list,
        label=[],
        title="Uversky plot",
        legendOn=True,
        xLim=1,
        yLim=1,
        fontSize=10,
        getFig=False):
    """
    Function to show multiple points on a Uversky plots

    """

    # Create a plotting object
    initial_plottingObject = multiple_plot(
        mean_net_charge_list, hydropathy_list, label, fontSize)
    finalized_plottingObject = finalize_uversky(
        initial_plottingObject, legendOn, title, xLim, yLim)

    # show the plot, or return the matplotlib fig object
    if getFig:
        return finalized_plottingObject
    else:
        finalized_plottingObject.show()


#...................................................................................#
def save_multiple_uverskyPlot(
        hydropathy_list,
        mean_net_charge_list,
        filename,
        label=[],
        title="Uversky plot",
        legendOn=True,
        xLim=1,
        yLim=1,
        fontSize=10,
        saveFormat='png'):

    # Create a plotting object
    initial_plottingObject = multiple_plot(
        mean_net_charge_list, hydropathy_list, label, fontSize)
    finalized_plottingObject = finalize_uversky(
        initial_plottingObject, legendOn, title, xLim, yLim)

    # save the plot and then close the matloblib plotting object
    if saveFormat == 'png':
        finalized_plottingObject.savefig(filename, format=saveFormat, dpi=200)
    else:
        finalized_plottingObject.savefig(filename, format=saveFormat)

    plt.close()


## <><><><><><><><><><><><><><><><><><><><><><><><><><><><><><> ##
##                                                              ##
##   INTERNAL FUNCTIONS FOR Uversky/Phase diagram plots         ##
##                                                              ##
## <><><><><><><><><><><><><><><><><><><><><><><><><><><><><><> ##

#...................................................................................#
def single_plot(x, y, label="", fontSize=10):
    """
    Internal function for creating a single sequence MATPLOTLIB object which can
    either then be saved or be displayed.

    INPUT:
    x          | x-coordinate to be plotted
    y          | y-coordinate to be plotted
    label      | label associated with that point (DEFAULT = "")
    fontSize  | size of label font (DEFAULT = 10)

    OUTPLOT:
    matplotlib object which can be plotted or saved

    """

    # set the x and y position of the mark
    x = float(x)
    y = float(y)

    # draw the actual plot
    plt.scatter(x, y, s=50, marker='o', color='Black', zorder=5)

    # if no label return plt object
    if label == "":
        return plt

    # we scale the location of the label in the event of extreme values
    if x > 0.8:
        x_lab = x - (0.01 * len(label) + 0.03)
    else:
        x_lab = x + 0.01

    if y > 0.9:
        y_lab = y - 0.025
    else:
        y_lab = y + 0.01

    # add the plot label
    plt.annotate(label, xy=(x_lab, y_lab), fontsize=fontSize)

    return plt


#...................................................................................#
def multiple_plot(x_list, y_list, label_list, fontSize):
    """
    Internal function for creating multiple induvidual points on a single set of axes.
    This plot can then be saved or displayed

    NOTE: This function ASSUMES x_list and y_list are lists of floats OF THE SAME
    LENGTH - this should be checked before!!

    INPUT:
    x_list     | list of x values to plot
    y_list     | list of y values to plot
    label_list | list of labels to associate with the points
    fontSize  | size of label font (DEFAULT = 10)

    OUTPLOT:
    matplotlib object which can be plotted or saved

    """

    # if we have no labels then construct the empty label list
    if len(label_list) == 0:
        # construct an empty label list
        label_list = []
        for i in xrange(0, len(x_list)):
            label_list.append("")

    # check that the three lists are the same length
    if not (len(x_list) == len(y_list) == len(label_list)):
        raise PlottingException(
            "Unequal length of positive fraction list, negative fraction list, and label list")

    # plot all the points
    for x, y, label in zip(x_list, y_list, label_list):
        plt.scatter(x, y, s=10, marker='o', color='Black', zorder=2)
        plt.annotate(label, xy=(x, y + 0.01), fontsize=fontSize)

    # annotate, set it all up, and return the plot object
    return plt


#...................................................................................#
def finalize_DasPappu(plt, legendOn, title, xLim, yLim):
    """
    Common function which finalizes up a plot by drawing on the regions 1-5, adding
    the legend and title. Used by both single and multiple phasePlot function

    """

    # define the five regions by filling the plot
    alphaval = 1
    reg1, = plt.fill([0, 0, 0.25], [0, 0.25, 0],
                     color='Chartreuse', alpha=alphaval, zorder=1)
    reg2, = plt.fill([0, 0, 0.35, 0.25], [0.25, 0.35, 0, 0],
                     color='MediumSeaGreen', alpha=alphaval, zorder=1)
    reg3, = plt.fill([0, 0.325, 0.675, 0.35], [0.35, 0.675,
                                               0.325, 0], color='DarkGreen', alpha=alphaval, zorder=1)

    reg4, = plt.fill([0, 0, 0.325], [0.35, 1, 0.675],
                     color='Red', alpha=alphaval, zorder=1)

    reg5, = plt.fill([0.35, 0.675, 1], [0, 0.325, 0],
                     color='Blue', alpha=alphaval, zorder=1)

    # set the plot limits
    plt.xlim([0, xLim])
    plt.ylim([0, yLim])

    # label the axes and set the title
    axes_pro = FontProperties()
    axes_pro.set_size('large')
    axes_pro.set_weight('bold')
    plt.xlabel(
        'Fraction of positively charged residues',
        fontproperties=axes_pro)
    plt.ylabel(
        'Fraction of negatively charged residues',
        fontproperties=axes_pro)

    # update the font property for the title
    axes_pro.set_size('x-large')
    plt.title(title, fontproperties=axes_pro)

    # if we the legend is on add the annotation
    if legendOn:

        # create and set set legend font options
        fontP = FontProperties()
        fontP.set_size('small')
        plt.legend([reg1, reg2, reg3, reg4, reg5],
                   ['Weak polyampholytes & polyelectrolytes:\nGlobules & tadpoles',
                    'Janus sequences:\nCollapsed or expanded - context dependent',
                    'Strong polyampholytes:\nCoils, hairpins, & chimeras',
                    'Negatively charged strong polyelectrolytes:\nSwollen coils',
                    'Positively charged strong polyelectrolytes:\nSwollen coils'],
                   prop=fontP)
    return plt


#...................................................................................#
def phaseplot_validate(fp, fn):
    """
    Validate if the fp and fn are resonable for making the diagram of states
    phase plots

    """

    # first check we can convert these guys into floating point numbers
    try:
        fp = float(fp)
    except ValueError as e:
        raise PlottingException(
            "Unable to convert " +
            str(fp) +
            " into a float")

    try:
        fn = float(fn)
    except ValueError as e:
        raise PlottingException(
            "Unable to convert " +
            str(fn) +
            " into a float")

    # next check they're both between 0 and 1
    if (fp < 0) or (fp > 1):
        raise PlottingException(
            "Fraction of positive residues outside of appropriate range [" + str(fp) + "]")
    if (fn < 0) or (fn > 1):
        raise PlottingException(
            "Fraction of positive residues outside of appropriate range [" + str(fn) + "]")

    # if we get here everything looks OK!


#...................................................................................#
def finalize_uversky(plt, legendOn, title, xLim, yLim):
    """
    Common function which finalizes up a plot by drawing on the regions 1-5, adding
    the legend and title. Used by both single and multiple phasePlot function

    """

    # define the five regions by filling the plot
    alphaval = 0.15

    # folded region
    reg1, = plt.fill([0.0, 0, 0.772], [1, 0.413, 1],
                     color='Chartreuse', alpha=0.25, zorder=1)

    # unfolded region
    reg2, = plt.fill([0, 0, 0.772, 1, 1],
                     [0, 0.413, 1, 1, 0],
                     color='Red',
                     alpha=0.15,
                     zorder=1)

    # set the plot limits
    plt.xlim([0, xLim])
    plt.ylim([0, yLim])

    # label the axes and set the title
    axes_pro = FontProperties()
    axes_pro.set_size('large')
    axes_pro.set_weight('bold')
    plt.xlabel('Mean net charge', fontproperties=axes_pro)
    plt.ylabel('Mean hydropathy <H>', fontproperties=axes_pro)

    # update the font property for the title
    axes_pro.set_size('x-large')
    plt.title(title, fontproperties=axes_pro)

    # if we the legend is on add the annotation
    if legendOn:

        # create and set set legend font options
        fontP = FontProperties()
        fontP.set_size('small')
        plt.legend([reg1, reg2], ['Folded proteins',
                                  'Natively unfolded'], prop=fontP)

    return plt


# <><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><>
#                           Linear sequence plots
# <><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><>

#...................................................................................#
def save_linearNCPR(SeqObj, blobLen, filename, saveFormat='png'):
    save_linearplot(build_NCPR_plot, SeqObj, blobLen, filename, saveFormat)


#...................................................................................#
def save_linearFCR(SeqObj, blobLen, filename, saveFormat='png'):
    save_linearplot(build_FCR_plot, SeqObj, blobLen, filename, saveFormat)


#...................................................................................#
def save_linearSigma(SeqObj, blobLen, filename, saveFormat='png'):
    save_linearplot(build_sigma_plot, SeqObj, blobLen, filename, saveFormat)


#...................................................................................#
def save_linearHydropathy(SeqObj, blobLen, filename, saveFormat='png'):
    save_linearplots(build_hydropathy_plot, SeqObj, blobLen, filename, saveFormat)

#...................................................................................#
def save_linearComplexity(complexityVector, complexityType, seqlen, filename, saveFormat='png'): 
    # Different structure to the others because here we have to pass a built complexity
    # vector into this function from the SeqObj
    

    plt = show_linearComplexity(complexityVector, complexityType, seqlen, getFig=True)
    
    if saveFormat == 'png':
        plt.savefig(filename, format=saveFormat, dpi=200)
    else:
        plt.savefig(filename, format=saveFormat)

    plt.close()

#...................................................................................#
def show_linearNCPR(SeqObj, blobLen, getFig=False):
    if getFig:
        return show_linearplot(build_NCPR_plot, SeqObj, blobLen, getFig)
    else:
        show_linearplot(build_NCPR_plot, SeqObj, blobLen, getFig)

#...................................................................................#
def show_linearFCR(SeqObj, blobLen, getFig=False):
    if getFig:
        return show_linearplot(build_FCR_plot, SeqObj, blobLen, getFig)
    else:
        show_linearplot(build_FCR_plot, SeqObj, blobLen, getFig)

#...................................................................................#
def show_linearSigma(SeqObj, blobLen, getFig=False):

    if getFig:
        return show_linearplot(build_sigma_plot, SeqObj, blobLen, getFig)
    else:
        show_linearplot(build_sigma_plot, SeqObj, blobLen, getFig)

#...................................................................................#
def show_linearHydropathy(SeqObj, blobLen, getFig=False):

    if getFig:
        return show_linearplot(build_hydropathy_plot, SeqObj, blobLen)
    else:
        show_linearplot(build_hydropathy_plot, SeqObj, blobLen)

#...................................................................................#
def show_linearComplexity(complexityVector, complexityType, seqlen, getFig=False):
    """
    The complexity plotting functions opperate outside of the general linear sequence
    framework as there are types of options/behaviours specific enough to the
    complexity plots that trying to shoe-horn them into the existing code would not
    be a good design decision.


    """
    
    # first generate the bar-plot and save the list of bars
    barlist = plt.bar(complexityVector[0,:], 
                      complexityVector[1,:],
                      width=1,
                      linewidth=1.0,
                      edgecolor='k',
                      color='#A8A8A8')

    # set the limits
    plt.ylim([0,1])
    plt.xlim([1, seqlen])

    # set the font properties
    axes_pro = FontProperties()
    axes_pro.set_size('large')
    axes_pro.set_weight('bold')

    # set the axis labels
    plt.xlabel('Residue', fontproperties=axes_pro)
    plt.ylabel('Complexity', fontproperties=axes_pro)
    
    # set the title (i.e. what type of complexity was calculated)
    axes_pro.set_size('x-large')
    if complexityType == 'WF':
        title='Wooton-Federhen complexity'
    elif complexityType == 'LC':
        title='Linguistic complexity'
    elif complexityType == 'LZW':
        title='Lempel-Ziv-Welch complexity'
    else:
        raise PlottingException('Unexpected complexity type passed - should never happen')

    plt.title(title, fontproperties=axes_pro)
    
    # finally either show the plot or return the plt object
    if getFig:
        return plt
    else:
        plt.show()


##
# Functions below allow construction of the various Linear Sequence Plots - i.e. should 
# be considered internal to this file
##

#...................................................................................#
def __build_linear_plot(
        data,
        title="",
        xlabel="Blob index",
        ylabel="",
        ylimits=[0,1],
        hline=None,
        setPositiveNegativeBars=False):
    """
    Internal function which expects data to be a Nx2 matrix (np.vstack) where column 1
    is the x values and column 2 is the y values. It also assumes the Y values
    are scaled to between 0 and 1 (so Y-axis limits are 0 and 1)

    hline defines a list of horizontal lines = so [0.2,-0.2] would draw horizontal lines


    """

    # plot the data
    barlist = plt.bar( data[0,:], 
                       data[1,:],
                       width=1,
                       linewidth=1.1,
                       edgecolor='k',
                       color='#A8A8A8')

    # this is really inefficient but means we have a consistent
    #
    if setPositiveNegativeBars:
        for bar in xrange(0, len(barlist)):
            if data[1, bar] < 0:
                barlist[bar].set_color('r')
                barlist[bar].set_edgecolor('k')
            else:
                barlist[bar].set_color('b')
                barlist[bar].set_edgecolor('k')

        # draw mao lines
        plt.plot([0, len(barlist)], [0.26, 0.26],
                 color='k', linewidth=1.5, linestyle="--")
        plt.plot([0, len(barlist)], [-0.26, -0.26],
                 color='k', linewidth=1.5, linestyle="--")

    # set Y lims
    plt.ylim(ylimits)
    plt.xlim([1, len(data[0, :])])

    axes_pro = FontProperties()
    axes_pro.set_size('large')
    axes_pro.set_weight('bold')

    # label
    plt.xlabel(xlabel, fontproperties=axes_pro)
    plt.ylabel(ylabel, fontproperties=axes_pro)

    axes_pro.set_size('x-large')
    plt.title(title, fontproperties=axes_pro)

    # return plot object
    return plt


#...................................................................................#
def build_NCPR_plot(SeqObj, blobLen):
    """
    Function which returns a matplotlib.plt object ready for saving/plotting
    of the NCPR along your sequence divided into blobs of some size
    """

    try:
        plt = __build_linear_plot(SeqObj.linearDistOfNCPR(blobLen),
                                  title='NCPR distribution (blob ' + str(int(blobLen)) + ')',
                                  ylabel='NCPR', ylimits=[-1, 1], setPositiveNegativeBars=True)
    except PlottingException:
        raise PlottingException(
            "NCPR plot construction requires Sequence object")

    return plt

#...................................................................................#


def build_FCR_plot(SeqObj, blobLen):
    """
    Function which returns a matplotlib.plt object ready for saving/plotting
    of the FCR along your sequence divided into blobs of some size
    """

    try:
        plt = __build_linear_plot(SeqObj.linearDistOfFCR(blobLen),
                                  title='FCR distribution (blob ' + str(int(blobLen)) + ')',
                                  ylabel='FCR')
    except PlottingException:
        raise PlottingException(
            "FCR plot construction requires Sequence object")

    return plt


#...................................................................................#
def build_sigma_plot(SeqObj, blobLen):
    """
    Function which returns a matplotlib.plt object ready for saving/plotting
    of the sigma along your sequence divided into blobs of some size
    """

    try:
        plt = __build_linear_plot(SeqObj.linearDistOfSigma(blobLen),
                                  title='Sigma distribution (blob ' + str(int(blobLen)) + ')',
                                  ylabel='Sigma')

    except AttributeError as e:
        raise PlottingException(
            "Sigma plot construction requires Sequence object")

    return plt

#...................................................................................#


def build_hydropathy_plot(SeqObj, blobLen):
    """
    Function which returns a matplotlib.plt object ready for saving/plotting
    of the Uversky-hydropathy along your sequence divided into blobs of some size
    """

    try:
        plt = __build_linear_plot(SeqObj.linearDistOfHydropathy(blobLen),
                                  title='Hydropathy distribution (blob ' + str(int(blobLen)) + ')',
                                  ylabel='Hydropathy')

    except AttributeError as e:
        raise PlottingException(
            "Hydropathy plot construction requires Sequence object")

    return plt


#...................................................................................#
def save_linearplot(build_fun, SeqObj, blobLen, filename, saveFormat='png'):
    """
    Internal function which builds and saves a linear sequence plot

    build_fun is the linear sequence building function we're using
    """

    plt = build_fun(SeqObj, blobLen)
    
    if saveFormat == 'png':
        plt.savefig(filename, format=saveFormat, dpi=200)
    else:
        plt.savefig(filename, format=saveFormat)

    plt.close()


#...................................................................................#
def show_linearplot(build_fun, SeqObj, blobLen, getFig=False):
    """
    Internal function which builds and shows a linear sequence plot

    build_fun is the linear sequence building function we're using
    """

    plt = build_fun(SeqObj, blobLen)
    if getFig:
        return plt
    else:
        plt.show()



#...................................................................................#
def save_local_composition_plot(residue_number, density_vectors, legend_color, legend_names, filename, saveFormat='png', max_val=-1, line_thickness=[], title='', plot_data=False):
    """
    Unlike the other plotting functions, the save_local_composition_plot is the only function for plotting local linear
    composition. It doesn't take advantage of the other builder functions, but instead exists as its own stand alone function.

    Arguments are as follows:

    residue_number - vector with 1-n numbering the residues

    density_vector - x by n np matrix where there are n values for each of the
                     n residues for x different residue groups

    legend_color - list of matplotlib compatible colors (must be the same length
                   as the number of groups)

    legend_names = list of names to assign for each group type (must be the same length
                   as the number of groups)

    filename - name of file to be saved

    save_format - format to be saved (default = 'png', 'pdf' also accepted)

    max_val - default maximum density value for residue group density (default =-1,
              which means the optimum value for dynamic range is calculated automatically
        

    """


    # determine the number of different groups
    n_groups = density_vectors.shape[0]

    # correct for bad starting values
    if max_val > 1 or max_val < 0:
        max_val = -1
        
    # determine the maximum density of any of the groups (note if a user-defined max_val is provided this
    # is skipped
    if max_val == -1:
        # this means max is 0.1 above max value OR 1.0, whicever is smaller
        max_val = min(np.max(density_vectors)+0.1, 1.0)

    # do some sanity checking to make sure all our ducks are in a row before we go all in
    if n_groups != len(legend_names):
        raise PlottingException('Mismatch in the number of of groups provided and the number of groups named for figure legend')

    # sanity check groups and legend color vector
    if  n_groups != len(legend_color):
        raise PlottingException('Mismatch in the number of of groups provided and the number of groups defined for line colors')
            

    ##
    ## The following section sets the figure size and font size, and is kind of an empyrical hack to make
    ## the generated figure look good and legible regardless of the number of residues. Works well!

    # if we're working with a sequence < 250 points
    if len(residue_number) < 250:
        width  = 20.0
        height = 12.5
        fs1    = 22
        fs2    = 28
        if len(line_thickness) == 0:
            for i in xrange(0, n_groups):
                line_thickness.append(1.5)

    # if we're working with a sequence > 250 points
    else:
        width  = 0.075*len(residue_number)
        height = 0.04*len(residue_number)
        fs1    = 0.08*len(residue_number)
        fs2    = 0.10*len(residue_number)
        if len(line_thickness) == 0:
            for i in xrange(0, n_groups):
                line_thickness.append(3.0)
        
    num_res = len(residue_number)
    
    handles_vector = []
    #fig = plt.figure(
    plt.figure(figsize=(width, height))

    font = {'family' : 'Bitstream Vera Sans',
            'weight' : 'normal',
            'size'   : fs1}

    matplotlib.rc('font', **font)

    
    for i in xrange(0, n_groups):
        
        # fit the density associated with group
        fx = UnivariateSpline(residue_number, density_vectors[i], s=1)

        if plot_data:
            h, = plt.plot(residue_number, density_vectors[i], color=legend_color[i], label=legend_names[i], linewidth=(line_thickness[i]*0.5))

        h, = plt.plot(residue_number, fx(residue_number), color=legend_color[i], label=legend_names[i], linewidth=line_thickness[i])
        handles_vector.append(h)


    # set the legend and bounding box positions so everything looks nice
    ax = plt.subplot(111)
    box = ax.get_position()
    ax.set_position([box.x0, box.y0 + box.height * 0.1, box.width, box.height * 0.9])                     
    plt.legend(handles=handles_vector, loc='upper center', bbox_to_anchor=(0.5, -0.05), fancybox=True, shadow=True, ncol=4)
               
    # set some asthetics stuff to make it look nice
    axes = plt.gca()
    axes.set_ylim([0,max_val])
    axes.set_xlim([1,num_res])
    plt.ylabel('Local Amino Acid Density', fontsize=fs2)

    # if we provided a title set that badboy
    if len(title) > 0:        
        plt.title(title, fontsize=fs2, loc='left')
    
    plt.savefig(filename, format=saveFormat)
    plt.close()




        

    



"""
   !--------------------------------------------------------------------------!
   ! LICENSE INFO:                                                            !
   !--------------------------------------------------------------------------!
   !    This file is part of localCIDER.                                      !
   !                                                                          !
   !    Version 0.1.9                                                         !
   !                                                                          !
   !    Copyright (C) 2014 - 2016                                             !
   !    The localCIDER development team (current and former contributors)     !
   !    Alex Holehouse, James Ahad, Rahul K. Das.                             !
   !                                                                          !
   !    localCIDER was developed in the lab of Rohit Pappu at Washington      !
   !    University in St. Louis. Please see the website for citation          !
   !    information:                                                          !
   !                                                                          !
   !    http://pappulab.github.io/LocalCIDER/                                 !
   !                                                                          !
   !    For more information please see the Pappu lab website:                !
   !                                                                          !
   !    http://pappulab.wustl.edu/                                            !
   !                                                                          !
   !    LocalCIDER is free software: you can redistribute it and/or modify    !
   !    it under the terms of the GNU General Public License as published by  !
   !    the Free Software Foundation, either version 3 of the License, or     !
   !    (at your option) any later version.                                   !
   !                                                                          !
   !    LocalCIDER is distributed in the hope that it will be useful,         !
   !    but WITHOUT ANY WARRANTY; without even the implied warranty of        !
   !    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         !
   !    GNU General Public License for more details.                          !
   !                                                                          !
   !    You should have received a copy of the GNU General Public License     !
   !    along with LocalCIDER.  If not, see <http://www.gnu.org/licenses/>.   !
   !--------------------------------------------------------------------------!
   ! AUTHORSHIP INFO:                                                         !
   !--------------------------------------------------------------------------!
   !                                                                          !
   ! MAIN AUTHOR:   Mary Richardson and Alex Holehouse                        !
   !                                                                          !
   !--------------------------------------------------------------------------!


   File Description:
   ================

"""
import zlib
import numpy as np
import math
from data.highComplexitySequences import maxComplexity

from localciderExceptions import SequenceException, SequenceComplexityException
from data.aminoacids import TWENTY_AAs


class SequenceComplexity:
    """
    Dedicated class to dealing with sequence complexity issues.

    This is a stateless class, but contains all the logic for carrying out
    sequence complexity calculations.

    """

    def __init__(self):
        pass



    #...................................................................................#
    def Zlib_compressed_complexity(self, sequence):
        """ Return the normalized sequence complexity, where normalization
            occurs relative to a random string of the same length.

            Note that we are explicitly defining sequence complexity based on the ability
            to minimally encode that specific sequence.
        """

        if len(sequence) > 1000:
            raise SequenceException(
                "Currently only sequences less than 1000 residues are subject to sequence complexity analysis - also, this feature is actually not officially out yet!")

        return float(len(zlib.compress(sequence))) / \
            maxComplexity[len(sequence)]



    #...................................................................................#
    def reduce_alphabet(self, sequence, alphabetSize=20, userAlphabet={}):
        """
        Converts your sequence to either one of the 11 predefined alphabets (defined based
        on size) or allows you to define your own alphabet using the userAlphabet dictionary.

        userAlphabet is checked for sanity (i.e. that it can deal with all 20 amino acids)

        Predefined alphabets are shown below

        two      - [(LVIMCAGSTPFYW), (EDNQKRH)]
        three    - [(LVIMCAGSTP), (FYW), (EDNQKRH)]
        four     - [(LVIMC), (AGSTP), (FYW), (EDNQKRH)]
        five     - [(LVIMC), (ASGTP), (FYW), (EDNQ), (KRH)]
        six      - [(LVIM), (ASGT), (PHC), (FYW), (EDNQ), (KR)]
        eight    - [(LVIMC), (AG), (ST), (P), (FYW), (EDNQ), (KR), (H)]
        ten      - [(LVIM), (C), (A), (G), (ST), (P), (FYW), (EDNQ), (KR), (H)]
        eleven   - [(LVIM), (C), (A), (G), (ST), (P), (FYW), (ED), (NQ), (KR), (H)]
        twelve   - [(LVIM), (C), (A), (G), (ST), (P), (FY), (W), (EQ), (DN), (KR), (H)]
        fifteen  - [(LVIM), (C), (A), (G), (S), (T), (P), (FY), (W), (E), (Q), (D), (N), (KR), (H)]
        eighteen - [(LM), (VI), (C), (A), (G), (S), (T), (P), (F), (Y), (W), (E), (D), (N), (Q), (K), (R), (H)]
        twenty   - all twenty!


        """

        two    = ['L', 'E']
        three  = ['L', 'F', 'E']
        four   = ['L', 'A', 'F', 'E']
        five   = ['L', 'A', 'F', 'E', 'K']
        six    = ['L', 'A', 'P', 'F', 'E', 'K']
        eight  = ['L', 'A', 'S', 'P', 'F', 'E', 'K', 'H']
        ten    = ['L', 'C', 'A', 'G', 'S', 'P', 'F', 'E', 'K', 'H']
        eleven = ['L', 'C', 'A', 'G', 'S', 'P', 'F', 'E', 'K', 'H', 'Q']
        twelve = ['L', 'C', 'A', 'G', 'S', 'P', 'F', 'W', 'E', 'D', 'K', 'H']
        fifteen = [
            'L',
            'C',
            'A',
            'G',
            'S',
            'T',
            'P',
            'F',
            'W',
            'E',
            'Q',
            'D',
            'N',
            'K',
            'H']
        eighteen = ['L', 'V', 'C', 'A', 'G', 'S', 'T', 'P',
                    'F', 'Y', 'W', 'E', 'D', 'N', 'Q', 'K', 'R', 'H']
        twenty = [
            'R',
            'H',
            'K',
            'D',
            'E',
            'S',
            'T',
            'N',
            'Q',
            'C',
            'G',
            'P',
            'A',
            'I',
            'L',
            'M',
            'F',
            'W',
            'Y',
            'V']

        aa = []
        alphabet = []

        # >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
        # USER DEFINED ALPHABET
        #
        # Note that if we're using a user defined alphabet then we ignore the alphabetSize
        # parameter entirly

        if len(userAlphabet) > 0:
            if not isinstance(userAlphabet, dict):
                raise SequenceComplexityException(
                    'Invalid user alphabet supplied, must be a dictionary or dictionary derived object')

            # this is all sanity checking for our userdefined alphabet, but is important
            # because most of the other functions assume the sequence being
            # passed is valid
            for x in TWENTY_AAs:
                try:
                    converted = userAlphabet[x]
                except KeyError:
                    raise SequenceComplexityException(
                        'Invalid user alphabet supplied - does not allow mapping of amino acid %s' % x)

                if converted not in TWENTY_AAs:
                    raise SequenceComplexityException(
                        'Invalid user alphabet supplied - amino acid %s maps to %s, which is not a valid amino acid (must be upper case)' %
                        (x, converted))

            # build the reduced sequence
            for x in sequence:
                aa.append(userAlphabet[x])

            # finally build the alphabet
            alphabet = []
            for x in TWENTY_AAs:
                val = userAlphabet[x]
                if val not in alphabet:
                    alphabet.append(val)

            return ("".join(aa), alphabet)

        # >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
        # PREDEFINED DEFINED ALPHABET
        #
        # NOTE - we don't get here if we've used a userdefined alphabet
        # check the type of alphabetSize is OK
        try:
            alphabetSize = int(alphabetSize)
        except ValueError:
            raise SequenceComplexityException(
                "alphabetSize should be an number, or a string which can be converted into an integer")

        # check that one of the known alphabets is being used
        if alphabetSize not in [2, 3, 4, 5, 6, 8, 10, 11, 12, 15, 18, 20]:
            raise SequenceComplexityException(
                'Predefined alphabet sizes must be one of (2,3,4,5,6,8,10,11 12,15,18,20)\nGot [%i]'%alphabetSize)

        # Now we figure out which alphabet we're using and then generate a reduced resolution
        # sequence
        # 2: [(LVIMCAGSTPFYW), (EDNQKRH)]
        if (alphabetSize == 2):
            alphabet = two
            for x in sequence:
                if x in (
                        'L',
                        'V',
                        'I',
                        'M',
                        'C',
                        'A',
                        'G',
                        'S',
                        'T',
                        'P',
                        'F',
                        'Y',
                        'W'):
                    aa.append('L')
                else:
                    aa.append('E')

        # 3: [(LVIMCAGSTP), (FYW), (EDNQKRH)]
        elif (alphabetSize == 3):
            alphabet = three
            for x in sequence:
                if x in ('L', 'V', 'I', 'M', 'C', 'A', 'G', 'S', 'T', 'P'):
                    aa.append('L')
                elif x in ('F', 'Y', 'W'):
                    aa.append('F')
                else:
                    aa.append('E')

        # 4: [(LVIMC), (AGSTP), (FYW), (EDNQKRH)]
        elif (alphabetSize == 4):
            alphabet = four
            for x in sequence:
                if x in ('L', 'V', 'I', 'M', 'C'):
                    aa.append('L')
                elif x in ('A', 'G', 'S', 'T', 'P'):
                    aa.append('A')
                elif x in ('F', 'Y', 'W'):
                    aa.append('F')
                else:
                    aa.append('E')

        # 5: [(LVIMC), (ASGTP), (FYW), (EDNQ), (KRH)]
        elif (alphabetSize == 5):
            alphabet = five
            for x in sequence:
                if x in ('L', 'V', 'I', 'M', 'C'):
                    aa.append('L')
                elif x in ('A', 'S', 'G', 'T', 'P'):
                    aa.append('A')
                elif x in ('F', 'Y', 'W'):
                    aa.append('F')
                elif x in ('E', 'D', 'N', 'Q'):
                    aa.append('E')
                else:
                    aa.append('K')

        # 6: [(LVIM), (ASGT), (PHC), (FYW), (EDNQ), (KR)]
        elif (alphabetSize == 6):
            alphabet = six
            for x in sequence:
                if x in ('L', 'V', 'I', 'M'):
                    aa.append('L')
                elif x in ('A', 'S', 'G', 'T'):
                    aa.append('A')
                elif x in ('P', 'H', 'C'):
                    aa.append('P')
                elif x in ('F', 'Y', 'W'):
                    aa.append('F')
                elif x in ('E', 'D', 'N', 'Q'):
                    aa.append('E')
                else:
                    aa.append('K')

        # 8: [(LVIMC), (AG), (ST), (P), (FYW), (EDNQ), (KR), (H)]
        elif (alphabetSize == 8):
            alphabet = eight
            for x in sequence:
                if x in ('L', 'V', 'I', 'M', 'C'):
                    aa.append('L')
                elif x in ('A', 'G'):
                    aa.append('A')
                elif x in ('S', 'T'):
                    aa.append('S')
                elif x in ('F', 'Y', 'W'):
                    aa.append('F')
                elif x in ('E', 'D', 'N', 'Q'):
                    aa.append('E')
                elif x in ('H'):
                    aa.append('H')
                elif x in ('P'):
                    aa.append('P')
                else:
                    aa.append('K')

        # 10: [(LVIM), (C), (A), (G), (ST), (P), (FYW), (EDNQ), (KR), (H)]
        elif (alphabetSize == 10):
            alphabet = ten
            for x in sequence:
                if x in ('L', 'V', 'I', 'M'):
                    aa.append('L')
                elif x in ('S', 'T'):
                    aa.append('S')
                elif x in ('F', 'Y', 'W'):
                    aa.append('F')
                elif x in ('E', 'D', 'N', 'Q'):
                    aa.append('E')
                elif x in ('K', 'R'):
                    aa.append('K')
                else:
                    aa.append(x)

        # 11: [(LVIM), (C), (A), (G), (ST), (P), (FYW), (ED), (NQ), (KR), (H)]
        elif (alphabetSize == 11):
            alphabet = eleven
            for x in sequence:
                if x in ('L', 'V', 'I', 'M'):
                    aa.append('L')
                elif x in ('S', 'T'):
                    aa.append('S')
                elif x in ('F', 'Y', 'W'):
                    aa.append('F')
                elif x in ('E', 'D'):
                    aa.append('E')
                elif x in ('N', 'Q'):
                    aa.append('Q')
                elif x in ('K', 'R'):
                    aa.append('K')
                else:
                    aa.append(x)

        # 12: [(LVIM), (C), (A), (G), (ST), (P), (FY), (W), (EQ), (DN), (KR),
        # (H)]
        elif (alphabetSize == 12):
            alphabet = twelve
            for x in sequence:
                if x in ('L', 'V', 'I', 'M'):
                    aa.append('L')
                elif x in ('S', 'T'):
                    aa.append('S')
                elif x in ('F', 'Y'):
                    aa.append('F')
                elif x in ('E', 'Q'):
                    aa.append('E')
                elif x in ('D', 'N'):
                    aa.append('D')
                elif x in ('K', 'R'):
                    aa.append('K')
                else:
                    aa.append(x)

        # 15: [(LVIM), (C), (A), (G), (S), (T), (P), (FY), (W), (E), (Q), (D),
        # (N), (KR), (H)]
        elif (alphabetSize == 15):
            alphabet = fifteen
            for x in sequence:
                if x in ('L', 'V', 'I', 'M'):
                    aa.append('L')
                elif x in ('F', 'Y'):
                    aa.append('F')
                elif x in ('K', 'R'):
                    aa.append('K')
                else:
                    aa.append(x)

        # 18: [(LM), (VI), (C), (A), (G), (S), (T), (P), (F), (Y), (W), (E),
        # (D), (N), (Q), (K), (R), (H)]
        elif (alphabetSize == 18):
            alphabet = eighteen
            for x in sequence:
                if x in ('L', 'M'):
                    aa.append('L')
                elif x in ('V', 'I'):
                    aa.append('V')
                else:
                    aa.append(x)
        elif (alphabetSize == 20):
            alphabet = twenty
            aa = sequence
        else:
            print (
                "ERROR: invalid aa alphabet selected, using 20-letter alphabet instead")
            alphabet = twenty
            aa = sequence

        return ("".join(aa), alphabet)
    
        

    #...................................................................................#
    ###########################################################
    # calculate Wootton-Federhen complexity
    ###########################################################
    def CWF(self, sequence, alphabet, windowSize, stepSize):
        """
        Function to calculate the Wootton-Federhen complexity

        Requires four parameters
        1) Amino acid sequence
        2) Alphabet being used
        3) windowSize for sliding window
        4) Stepsize for moving along the sliding window

        """

        # Initialize the current step
        step = 0

        CWF_array = []

        # for each position
        while (step <= len(sequence) - windowSize):

            # restart complexity calculation for this window
            CWF = 0

            # get the current window
            window = sequence[step:step + windowSize]

            # for every residue in the alphabet
            for x in alphabet:

                # p is between 0 and 1
                p = float(window.count(x)) / windowSize

                if p > 0:
                    CWF = p * (math.log(p, len(alphabet))) + CWF

            # store the complexity score for this window
            CWF_array.append(-CWF)
            step = step + stepSize  # increment the step

        return CWF_array



    #...................................................................................#
    ###########################################################
    # calculate linguistic complexity
    ###########################################################
    def LC(self, sequence, alphabet, windowSize, stepSize, wordSize):
        # Simple explanation of linguistic complexity
        #
        # Linguistic complexity is basically asking, given a window of some size,
        # lets figure out the number of unique 'words' of some (specific) size in
        # that window, and divide that by the maximum possible number of unique words
        # in the window. We do this for each window, to create a vectorial
        # representation of the complexity

        # the current step
        step = 0
        LC_array = []

        while (step <= len(sequence) - windowSize):

            # restart complexity calculation for this window
            LC = 0

            # reset the position for this window
            i = 0

            ngrams = set()

            ngram = ''

            # for each position in the window we step through with a word of
            # size wordSize
            for i in range(0, windowSize - wordSize):

                # extract the current ngram
                position = step + i

                ngram = ''.join(sequence[position:position + wordSize])

                # if this ngram is not already present in the set
                if ngram not in ngrams:
                    # add it to the ngrams set
                    ngrams.add(ngram)

            # size of ngrams set
            v = len(ngrams)

            # max possible number of words
            vmax = min(len(alphabet)**wordSize, windowSize - 1 + wordSize)

            # linguistic complexity associated with this window
            LC = float(v) / vmax

            # add this to an array of the complexity profile scores and
            # increment the step
            LC_array.append(LC)
            step = step + stepSize

        return LC_array



    #...................................................................................#
    ###########################################################
    # calculate Lempel-Ziv-Welch complexity
    ###########################################################
    def LZW(self, sequence, alphabet, windowSize, stepSize):
        # the current step
        step = 0

        LZW_array = []

        while (step <= len(sequence) - windowSize):
            LZW = 0  # restart complexity calculation for this window
            i = 0   # reset the position for this window
            w = ''  # reset the word

            ngrams = set()

            # for each position in the window
            for i in range(0, windowSize):
                position = step + i

                # if we find a word which is already in
                # the ngrams set, update w to that
                if (w + sequence[position] in ngrams):

                    w = sequence[position] + w
                    #w = w+sequence[position]

                # else we found a new word
                else:
                    ngrams.add(w + sequence[position])
                    w = sequence[position]

            # size of ngrams set
            n = len(ngrams)

            if (windowSize > 0):
                LZW = float(n) / windowSize
                # add this to an array of the complexity profile scores
                LZW_array.append(LZW)
            step += stepSize         # increment the step

        return LZW_array



    #...................................................................................#
    def get_WF_complexity(
            self,
            sequence,
            alphabetSize=20,
            userAlphabet={},
            windowSize=10,
            stepSize=1):

        # reduce alphabet complexity
        (reduced_sequence, alphabet) = self.reduce_alphabet(
            sequence, alphabetSize, userAlphabet)

        complexity_vector = self.CWF(reduced_sequence, alphabet, windowSize, stepSize)

        return self.get_indexed_complexity_vector(complexity_vector, len(sequence))

        

    def get_LC_complexity(
            self,
            sequence,
            alphabetSize=20,
            userAlphabet={},
            windowSize=10,
            stepSize=1,
            wordSize=3):

        # reduce alphabet complexity
        (reduced_sequence, alphabet) = self.reduce_alphabet(
            sequence, alphabetSize, userAlphabet)

        complexity_vector =  self.LC(
            reduced_sequence,
            alphabet,
            windowSize,
            stepSize,
            wordSize)

        return self.get_indexed_complexity_vector(complexity_vector, len(sequence))

        
    def get_LZW_complexity(
            self,
            sequence,
            alphabetSize=20,
            userAlphabet={},
            windowSize=10,
            stepSize=1):

        # reduce alphabet complexity
        (reduced_sequence, alphabet) = self.reduce_alphabet(
            sequence, alphabetSize, userAlphabet)


        complexity_vector = self.LZW(reduced_sequence, alphabet, windowSize, stepSize)

        return self.get_indexed_complexity_vector(complexity_vector, len(sequence))


    def get_indexed_complexity_vector(self, complexity_vector, seq_len):
        """
        Takes a complexity vector of length N and returns a 2xN matrix containing
        the complexity vector and the corresponding residue positions distributed
        equally along the sequence of length seq_len

        """

        # having retrieved the vector we want to also return indicies which represent
        # an even distribution of points across the sequence (this is always going
        # to be the correct representation of the vector)
        complexity_vector_len = len(complexity_vector)
        spacing = seq_len/complexity_vector_len

        # based on the spacing calculate the necessary remainder (i.e. what's left
        # over..)
        remainder = (seq_len - (spacing*complexity_vector_len))

        if remainder % 2 == 0:
            flank_start = remainder/2
            flank_end = remainder/2
        else:
            flank_start = (remainder-1)/2
            flank_end = (remainder+1)/2

        # finally set the start and end positions as half-way through the spacing
        # unit + the appropriate flanking sequence offset from either the first
        # or last residue

        index_start = (flank_start+1) + spacing/2
        index_end   = ((seq_len + 1)-flank_end) + spacing/2
        indices = np.arange(index_start, index_end, spacing, dtype=int)

        return np.vstack((indices, complexity_vector))




"""
   !--------------------------------------------------------------------------!
   ! LICENSE INFO:                                                            !
   !--------------------------------------------------------------------------!
   !    This file is part of localCIDER.                                      !
   !                                                                          !
   !    Version 0.1.9                                                         !
   !                                                                          !
   !    Copyright (C) 2014 - 2016                                             !
   !    The localCIDER development team (current and former contributors)     !
   !    Alex Holehouse, James Ahad, Rahul K. Das.                             !
   !                                                                          !
   !    localCIDER was developed in the lab of Rohit Pappu at Washington      !
   !    University in St. Louis. Please see the website for citation          !
   !    information:                                                          !
   !                                                                          !
   !    http://pappulab.github.io/localCIDER/                                 !
   !                                                                          !
   !    For more information please see the Pappu lab website:                !
   !                                                                          !
   !    http://pappulab.wustl.edu/                                            !
   !                                                                          !
   !    localCIDER is free software: you can redistribute it and/or modify    !
   !    it under the terms of the GNU General Public License as published by  !
   !    the Free Software Foundation, either version 3 of the License, or     !
   !    (at your option) any later version.                                   !
   !                                                                          !
   !    localCIDER is distributed in the hope that it will be useful,         !
   !    but WITHOUT ANY WARRANTY; without even the implied warranty of        !
   !    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         !
   !    GNU General Public License for more details.                          !
   !                                                                          !
   !    You should have received a copy of the GNU General Public License     !
   !    along with localCIDER.  If not, see <http://www.gnu.org/licenses/>.   !
   !--------------------------------------------------------------------------!
   ! AUTHORSHIP INFO:                                                         !
   !--------------------------------------------------------------------------!
   !                                                                          !
   ! MAIN AUTHOR:   Alex Holehouse                                            !
   !                                                                          !
   !--------------------------------------------------------------------------!


   Unit tests for functions from backened/sequence.py


"""

import os.path
import time
import unittest
import random
from localcider.backend import sequence
import testTools


def roundflt(num):
    return float("%1.2f" % num)

class TestSequenceFunctions(unittest.TestCase):

    def setUp(self):
        self.SO_60 = sequence.Sequence('MEEPQSDPSVEPPLSQETFSDLWKLLPENNVLSPLPSQAMDDLMLSPDDIEQWFTEDPGP')
        self.SO_10 = sequence.Sequence('MEEPQSDPSV')


    def test_FCR(self):
        self.assertEqual(self.SO_60.FCR(), 0.25)
        self.assertEqual(self.SO_10.FCR(), 0.3)


    def test_NCPR(self):
        self.assertEqual(roundflt(self.SO_60.NCPR()), -0.22)
        self.assertEqual(roundflt(self.SO_10.NCPR()), -0.3)


    def test_fplus_fminus(self):
        self.assertEqual(roundflt(self.SO_60.Fminus()), 0.23)
        self.assertEqual(roundflt(self.SO_60.Fplus()), 0.02)

        self.assertEqual(roundflt(self.SO_10.Fminus()), 0.3)
        self.assertEqual(roundflt(self.SO_10.Fplus()),  0.0)

    def test_aminoAcidColorMap(self):
        
        # just code coverage
        CM = self.SO_10.aminoAcidColorMap
        self.assertEqual(len(CM), 20)

    def test_amino_acid_fraction(self):
        
        vals = {'A': 0.016666666666666666,
                'C': 0.0,
                'D': 0.11666666666666667,
                'E': 0.11666666666666667,
                'F': 0.03333333333333333,
                'G': 0.016666666666666666,
                'H': 0.0,
                'I': 0.016666666666666666,
                'K': 0.016666666666666666,
                'L': 0.13333333333333333,
                'M': 0.05,
                'N': 0.03333333333333333,
                'P': 0.16666666666666666,
                'Q': 0.06666666666666667,
                'R': 0.0,
                'S': 0.11666666666666667,
                'T': 0.03333333333333333,
                'V': 0.03333333333333333,
                'W': 0.03333333333333333,
                'Y': 0.0}

        fractions = self.SO_60.amino_acid_fraction()
        for i in vals:
            self.assertEqual(roundflt(fractions[i]), roundflt(vals[i]))


    def test_linear_vectors(self):
        
        # charge pattern vector
        self.assertEqual(len(self.SO_60.chargePattern), 60)

        self.assertEqual(len(self.SO_60.linearDistOfFCR(5)[0]), 60)
        self.assertEqual(len(self.SO_60.linearDistOfFCR(6)[0]), 60)

        self.assertEqual(len(self.SO_60.linearDistOfNCPR(5)[0]), 60)
        self.assertEqual(len(self.SO_60.linearDistOfNCPR(6)[0]), 60)

        self.assertEqual(len(self.SO_60.linearDistOfHydropathy(5)[0]), 60)
        self.assertEqual(len(self.SO_60.linearDistOfHydropathy(6)[0]), 60)

        self.assertEqual(len(self.SO_60.linearDistOfHydropathy_2(5)[0]), 60)
        self.assertEqual(len(self.SO_60.linearDistOfHydropathy_2(6)[0]), 60)

        self.assertEqual(len(self.SO_60.linearDistOfSigma(5)[0]), 60)
        self.assertEqual(len(self.SO_60.linearDistOfSigma(6)[0]), 60)
            
        






"""
   !--------------------------------------------------------------------------!
   ! LICENSE INFO:                                                            !
   !--------------------------------------------------------------------------!
   !    This file is part of localCIDER.                                      !
   !                                                                          !
   !    Version 0.1.9                                                         !
   !--------------------------------------------------------------------------!

   File Description:
   ================

   ***********************************************************************
   *                                                                     *
   * WARNING: SequencePermutants is still undergoing testing and will be *
   * released in version 0.2.0                                           *
   *                                                                     *
   ************************************************************************

   This file contains functionality for generating sequence permutants via
   the Wang-Landau algrorithm. There are two seperate operations which can
   be carried out here;

   1) Estimate the Kappa density of states over some region

   2) Create permutants of your sequence with various kappa values

   The general approach taken is that you create a sequencePermutant
   object with some sequence, and then call an operation on that object.


"""


class SequencePermutantException(Exception):
    pass


import numpy as np

from sequenceParameters import SequenceParameters
from backend.sequence import Sequence
from backend.seqfileparser import SequenceFileParser
from backend.backendtools import status_message, warn_notReadyYet

from backend import wang_landau


class SequencePermutants:

    #...................................................................................#
    def __init__(self, sequence="", sequenceFile=""):
        """
        sequencePermutant objects, like sequenceParameter objects, can be initialized by either an amino acid 
        sequence as a string, or a sequence file. Such sequence files can be FASTA files, or simply an amino 
        acid sequence in a text file.

        """

        warn_notReadyYet()

        # provide the flexibility to submit either a sequence
        # file or an actual sequence as a string
        if sequence == "" and sequenceFile == "":
            return None

        # if the sequence isn't empty constuct a local
        # sequence object using the sequence
        if not sequence == "":
            self.SeqObj = Sequence(sequence)
        else:
            parserMachine = SequenceFileParser()
            self.SeqObj = Sequence(parserMachine.parseSeqFile(sequenceFile))

        self.WL_ready = False

    #...................................................................................#
    def initializeWangLandauParameters(
            self,
            OUTDIR,
            frozen=set(
                []),
            nbins=10,
            binmin=0,
            binmax=1,
            flatchck=10000,
            flatcrit=0.7,
            convergence=np.exp(.000001),
            WL_type="NORMAL"):

        # initialize the Wang-Landau Machine
        self.WLM = wang_landau.WangLandauMachine(
            self.SeqObj.seq,
            OUTDIR,
            frozen,
            nbins,
            binmin,
            binmax,
            flatchck,
            flatcrit,
            convergence,
            WL_type)

        # set the ready flag
        self.WL_ready = True

    #...................................................................................#
    def run_wang_landau_DOS(self, binmin, minmax, numberofBins):
        # self.__readyCheck()
        warn_notReadyYet()

    #...................................................................................#
    def run_histogramZoom_DOS_estimation(self, nbins):
        warn_notReadyYet()

    #...................................................................................#
    def run_generatePermutants(self, numSequencesPerBin):
        warn_notReadyYet()

    #...................................................................................#
    def __readyCheck(self):
        if not self.WL_ready:
            raise SequencePermutantException(
                "Wang Landau not yet parameterized\nPLEASE RUN <initializeWangLandauParameters> BEFORE CALLING a run function")

    #...................................................................................#
    def get_permutant(self):

        # create a permutant Sequence object
        SO = self.SeqObj.full_shuffle([])

        # now create an 'empty' SequenceParameter object
        SP = SequenceParameters("A")

        # manually set the underlying Sequence object
        # and return
        SP.SeqObj = SO
        return SP





"""
   !--------------------------------------------------------------------------!
   ! LICENSE INFO:                                                            !
   !--------------------------------------------------------------------------!
   !    This file is part of localCIDER.                                      !
   !                                                                          !
   !    Version 0.1.9                                                         !
   !                                                                          !
   !    Copyright (C) 2014 - 2016                                             !
   !    The localCIDER development team (current and former contributors)     !
   !    Alex Holehouse, James Ahad, Rahul K. Das.                             !
   !                                                                          !
   !    localCIDER was developed in the lab of Rohit Pappu at Washington      !
   !    University in St. Louis. Please see the website for citation          !
   !    information:                                                          !
   !                                                                          !
   !    http://pappulab.github.io/localCIDER/                                 !
   !                                                                          !
   !    For more information please see the Pappu lab website:                !
   !                                                                          !
   !    http://pappulab.wustl.edu/                                            !
   !                                                                          !
   !    localCIDER is free software: you can redistribute it and/or modify    !
   !    it under the terms of the GNU General Public License as published by  !
   !    the Free Software Foundation, either version 3 of the License, or     !
   !    (at your option) any later version.                                   !
   !                                                                          !
   !    localCIDER is distributed in the hope that it will be useful,         !
   !    but WITHOUT ANY WARRANTY; without even the implied warranty of        !
   !    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         !
   !    GNU General Public License for more details.                          !
   !                                                                          !
   !    You should have received a copy of the GNU General Public License     !
   !    along with localCIDER.  If not, see <http://www.gnu.org/licenses/>.   !
   !--------------------------------------------------------------------------!
   ! AUTHORSHIP INFO:                                                         !
   !--------------------------------------------------------------------------!
   !                                                                          !
   ! MAIN AUTHOR:   Alex Holehouse                                            !
   !                                                                          !
   !--------------------------------------------------------------------------!


   Runner code for localcider tests. The various flags define which tests are run
   setting ALL to True means that all the tests will be run regardless of the 
   status of the other flags.

   This is set up to make it easy to quickly test specific components, rather
   than running the entire testing suite at once.

"""

import unittest
import __init__ as test


## TEST flags
ALL = True
sequenceParameters_test = True
plots_test = False
sequence_test = False
complexity_test = False

## plotting tests
if plots_test or ALL:
    suite = unittest.TestLoader().loadTestsFromTestCase(
        test.test_plots.TestPlotsFunctions)

    unittest.TextTestRunner(verbosity=2).run(suite)

## sequence parameter tests
if sequenceParameters_test or ALL:
    suite = unittest.TestLoader().loadTestsFromTestCase(
        test.test_sequenceParameters.TestSequenceParametersFunctions)

    unittest.TextTestRunner(verbosity=2).run(suite)

## sequence tests
if sequence_test or ALL:
    suite = unittest.TestLoader().loadTestsFromTestCase(
        test.test_sequence.TestSequenceFunctions)

    unittest.TextTestRunner(verbosity=2).run(suite)

## complexity tests
if complexity_test or ALL:
    suite = unittest.TestLoader().loadTestsFromTestCase(
        test.test_complexity.TestComplexityFunctions)

    unittest.TextTestRunner(verbosity=2).run(suite)





"""
   !--------------------------------------------------------------------------!
   ! LICENSE INFO:                                                            !
   !--------------------------------------------------------------------------!
   !    This file is part of localCIDER.                                      !
   !                                                                          !
   !    Version 0.1.9                                                         !
   !                                                                          !
   !    Copyright (C) 2014 - 2016                                             !
   !    The localCIDER development team (current and former contributors)     !
   !    Alex Holehouse, James Ahad, Rahul K. Das.                             !
   !                                                                          !
   !    localCIDER was developed in the lab of Rohit Pappu at Washington      !
   !    University in St. Louis. Please see the website for citation          !
   !    information:                                                          !
   !                                                                          !
   !    http://pappulab.github.io/localCIDER/                                 !
   !                                                                          !
   !    For more information please see the Pappu lab website:                !
   !                                                                          !
   !    http://pappulab.wustl.edu/                                            !
   !                                                                          !
   !    localCIDER is free software: you can redistribute it and/or modify    !
   !    it under the terms of the GNU General Public License as published by  !
   !    the Free Software Foundation, either version 3 of the License, or     !
   !    (at your option) any later version.                                   !
   !                                                                          !
   !    localCIDER is distributed in the hope that it will be useful,         !
   !    but WITHOUT ANY WARRANTY; without even the implied warranty of        !
   !    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         !
   !    GNU General Public License for more details.                          !
   !                                                                          !
   !    You should have received a copy of the GNU General Public License     !
   !    along with localCIDER.  If not, see <http://www.gnu.org/licenses/>.   !
   !--------------------------------------------------------------------------!
   ! AUTHORSHIP INFO:                                                         !
   !--------------------------------------------------------------------------!
   !                                                                          !
   ! MAIN AUTHOR:   Alex Holehouse                                            !
   !                                                                          !
   !--------------------------------------------------------------------------!

   File Description:
   ================

   keyfile parses a localCIDER keyfile for the more complex calculations. Right
   now this is just Wang-Landau based density of kappa state calculations, but
   there's no reason why more advanced analytical techniques can't be added in
   the future.

   With version 0.1.0 the keyfile functionality is redundant, as we don't provide
   stand-alone binaries. However, that is a main priority of version 0.3.0

"""


import os
import numpy as np
from backendtools import warning_message, status_message
from seqfileparser import SequenceFileParser

from localciderExceptions import KeyFileException

#=======================================================
# Default values for non-crucial keyfile parameters
DEFAULT_VALS = {
    'BIN_MIN': 0.05,
    'BIN_MAX': 0.95,
    'NUMBER_OF_BINS': 20,
    'FLATCHECK_FREQ': 10000,
    'CONVERGENCE': np.exp(0.0001),
    'FLATNESS_CRITERION': 0.7,
    'WL_TYPE': 'ZOOM'}

# list of expected keywords
KEYWORD_LIST = [
    'SEQFILE',
    'OUTDIR',
    'FREEZE_FILE',
    'BIN_MIN',
    'BIN_MAX',
    'NUMBER_OF_BINS',
    'FLATCHECK_FREQ',
    'CONVERGENCE',
    'FLATNESS_CRITERION',
    'WL_TYPE']

# types of WL
WL_TYPES = ['NORMAL', 'ZOOM']

#======================================================


class KeyFile():
    """
        KeyFile objects take a filename and do all the reuqired keyfile parsing during initializtaion
        creating a single object from which you can get any WL keyfile parameters of interest
    """

    # SEQUENCE_FILE  - file containing sequence
    # FREEZE_FILE    - file defining regions to hold constant
    # BIN_START      - starting bin value (optional)
    # BIN_END        - ending bin value (optional)
    # NUMBER_OF BINS - number of bins
    # OUTDIR         - directory for output to be written to
    # WL             -

    #...................................................................................#
    def __init__(self, filename):
        """
           Initializer defines the keywords we expect to see, and takes a filename and parses it

        """

        # initialize and zero out the keywords in the object dictionary
        self.KEYWORDS = {}
        for i in KEYWORD_LIST:
            self.KEYWORDS[i] = ''

        # parse the keyfile
        self.parse_keyfile(filename)

    #...................................................................................#
    def parse_keyfile(self, filename):
        """
        Function which takes a filename and parses it into the keyfile object for easy
        interaction with the file's content
        """

        status_message("Parsing keyfile...")
        status_message("---------------------------------------")

        SeqFileParser = SequenceFileParser()  # create a sequence file parsing object

        # read file to end
        with open(filename) as filehandle:
            content = filehandle.readlines()

        # [PHASE 1 START]
        # PARSE THE KEYFILE
        for line in content:
            line = line.strip()

            # if empty line
            if len(line) == 0:
                continue

            # comments in the keyfile
            if line[0] == "#":
                continue

            # if inline comment kill everything after the comment
            # character
            if len(line.split("#")) > 1:
                line = line.split("#")[0]

            # finally remove any other trailing whitespace
            line = line.strip()

            # split the remaining by whitespace
            line_list = line.split(" ")

            # now cycle over the first value in the whitelist splitted lits
            # and check if it matches one of the predefined KEYWORDS
            if line_list[0].strip() in self.KEYWORDS:

                # if we find a keyword and there's a single string after the keyword load it into the
                # KEYWORDS dictionary (i.e. this is what we expect!)
                if len(line_list) == 2:
                    self.KEYWORDS[line_list[0].strip()] = line_list[1].strip()
                # there was more than one whitespace seperated string after the
                # keyword - we basically fail at this
                else:
                    raise KeyFileException("Error: Found keyword " +
                                           str(line_list[0].strip()) +
                                           " but unable to parse associated value")
            else:
                warning_message(
                    "Found unexpected keyword [" + str(line_list[0].strip()) + "] - ignorning...")

        # Now add default for the sequene, which will hopefully be set in the
        # next section by parsing the sequencefile
        self.KEYWORDS["SEQUENCE"] = ""
        # [PHASE 1 END]

        status_message("---------------------------------------")
        status_message("Keyfile parsed!\n")
        status_message("Validating keyfile contents")
        status_message("---------------------------------------")

        # Having parsed the keyfile we now validate the keyfile so
        # we don't have to worry about validation later on

        # VALIDATE the parsed values
        # [PHASE 2 START]
        #

        for keyword in KEYWORD_LIST:
            # extract the value associated with each keyword in turn
            value = self.KEYWORDS[keyword]

            ##
            # SEQUENCE FILE VALIDATION AND PARSING
            ##
            if keyword == "SEQFILE":
                if value == "":
                    raise KeyFileException(
                        "ERROR: No sequence file provided in keyfile (expecting keyword [SEQFILE])")
                else:
                    if not os.path.isfile(value):
                        raise KeyFileException(
                            "Expected " + str(value) + " to be file")

                    # if its a file lets try and extract a sequence from it!
                    self.KEYWORDS[
                        "SEQUENCE"] = SeqFileParser.parseSeqFile(value)

                    # if we get here we *should* now have a sequence...
                    if self.KEYWORDS["SEQUENCE"] == "" or self.KEYWORDS[
                            "SEQUENCE"] is None:
                        raise KeyFileException(
                            "ERROR: No sequence was parsed from the sequence file...")

            ##
            # OUTPUT DIRECTORY VALIDATION
            ##
            elif keyword == "OUTDIR":
                if value == "":
                    raise KeyFileException(
                        "ERROR: No output directory provided in keyfile (expecting keyword [OUTDIR]")
                else:
                    # creates the output directory if it doesn't already exist
                    if not os.path.exists(value):
                        status_message(
                            "Creating output directory " + str(value))
                        try:
                            os.makedirs(value)
                        except OSError as e:
                            print "----------------------------"
                            print ""
                            print "ERROR Creating output directory - do you have permission to create the directory [" + str(value) + "]"
                            print ""
                            print "----------------------------"
                            raise e
                    # if it does exist raise a quick warning
                    else:
                        # check if its empty
                        if len(os.listdir(value)) > 0:
                            warning_message(
                                "Output directory exists already and is not empty [RISK OF OVERWRITING!]")
                        else:
                            pass  # empty directory already exists - brilliant!

            ##
            # FREEZE FILE VALIDATION
            ##
            elif keyword == "FREEZE_FILE":
                if value == "":
                    pass  # no freeze file, no problem
                else:
                    if not os.path.isfile(value):
                        raise KeyFileException(
                            "Expected " + str(value) + " to be file")
                    status_message("Using freeze file")
                    self.KEYWORDS[keyword] = value

            ##
            # WL TYPE
            ##
            elif keyword == "WL_TYPE":
                if value == "":
                    self.KEYWORDS[keyword] = DEFAULT_VALS[keyword]
                    status_message(
                        "Setting WL type to default [" + str(DEFAULT_VALS[keyword]) + "]")
                else:
                    if value in WL_TYPES:
                        self.KEYWORDS[keyword] = value
                        status_message(
                            "Setting WL type to keyfile defined [" + str(value) + "]")
                    else:
                        raise KeyFileException(
                            "Unexpected WL algorithm type selected " + str(value) + " ")

                self.KEYWORDS[keyword] = value

            ##
            # SET NUMERIC VALUES
            ##

            elif keyword == "BIN_MIN":
                self.__set_numeric(keyword, value)

            elif keyword == "BIN_MAX":
                self.__set_numeric(keyword, value)

            elif keyword == "NUMBER_OF_BINS":
                self.__set_numeric(keyword, value)

            elif keyword == "FLATCHECK_FREQ":
                self.__set_numeric(keyword, value)

            elif keyword == "CONVERGENCE":
                self.__set_numeric(keyword, value)

            elif keyword == "FLATNESS_CRITERION":
                self.__set_numeric(keyword, value)

            else:
                raise KeyFileException("SHOULD NOT BE GETTING HERE...")

    #...................................................................................#
    def __set_numeric(self, keyword, value):
        """
           Function which sets the KEYWORDS dictionary $keyword value to $value if $value can be treated
           as a numerical value or uses the default if it wasn't set (BUT DOES NOT use the default
           if an in-parsable value was set - we want to know when things are going wrong, silent errors
           cost lives. Maybe.).
        """
        if value == "":
            status_message("Setting " + keyword +
                           " to default [" + str(DEFAULT_VALS[keyword]) + "]")
            self.KEYWORDS[keyword] = DEFAULT_VALS[keyword]
        else:
            try:
                float(value)
                status_message("Setting " + keyword +
                               " to keyfile defined [" + str(value) + "]")
                self.KEYWORDS[keyword] = value
            except ValueError:
                raise KeyFileException(
                    "\n\nERROR: Invalid value for " +
                    keyword +
                    " - unable to convert [" +
                    value +
                    "] into a number\n")

    #...................................................................................#
    def get_bin_min(self):
        return self.KEYWORDS["BIN_MIN"]

    #...................................................................................#
    def get_bin_max(self):
        return self.KEYWORDS["BIN_MAX"]

    #...................................................................................#
    def get_num_bins(self):
        return self.KEYWORDS["NUMBER_OF_BINS"]

    #...................................................................................#
    def get_flatcheck_freq(self):
        return self.KEYWORDS["FLATCHECK_FREQ"]

    #...................................................................................#
    def get_convergence(self):
        return self.KEYWORDS["CONVERGENCE"]

    #...................................................................................#
    def get_sequence(self):
        return self.KEYWORDS["SEQUENCE"]

    #...................................................................................#
    def get_freezefile(self):
        return self.KEYWORDS["FREEZE_FILE"]

    #...................................................................................#
    def get_outdir(self):
        return self.KEYWORDS["OUTDIR"]

    #...................................................................................#
    def get_flatness_criterion(self):
        return self.KEYWORDS["FLATNESS_CRITERION"]

    #...................................................................................#
    def get_WL_type(self):
        return self.KEYWORDS["WL_TYPE"]
